import os
from sys import platform
from unittest.mock import patch

import numpy as np
import pytest
from qtpy.QtCore import Qt

from napari._qt.widgets.qt_dims import QtDims
from napari.components import Dims


def test_creating_view(qtbot):
    """
    Test creating dims view.
    """
    ndim = 4
    dims = Dims(ndim=ndim)
    view = QtDims(dims)

    qtbot.addWidget(view)

    # Check that the dims model has been appended to the dims view
    assert view.dims == dims

    # Check the number of displayed sliders is two less than the number of
    # dimensions
    assert view.nsliders == view.dims.ndim
    assert np.sum(view._displayed_sliders) == view.dims.ndim - 2
    assert np.all(
        [
            s.isVisibleTo(view) == d
            for s, d in zip(
                view.slider_widgets, view._displayed_sliders, strict=False
            )
        ]
    )


def test_changing_ndim(qtbot):
    """
    Test changing the number of dimensions
    """
    ndim = 4
    view = QtDims(Dims(ndim=ndim))

    qtbot.addWidget(view)

    # Check that adding dimensions adds sliders
    view.dims.ndim = 5
    assert view.nsliders == view.dims.ndim
    assert np.sum(view._displayed_sliders) == view.dims.ndim - 2
    assert np.all(
        [
            s.isVisibleTo(view) == d
            for s, d in zip(
                view.slider_widgets, view._displayed_sliders, strict=False
            )
        ]
    )

    # Check that removing dimensions removes sliders
    view.dims.ndim = 2
    assert view.nsliders == view.dims.ndim
    assert np.sum(view._displayed_sliders) == view.dims.ndim - 2
    assert np.all(
        [
            s.isVisibleTo(view) == d
            for s, d in zip(
                view.slider_widgets, view._displayed_sliders, strict=False
            )
        ]
    )


def test_changing_focus(qtbot):
    """Test changing focus updates the dims.last_used prop."""
    # Initialize to 0th axis
    ndim = 2
    view = QtDims(Dims(ndim=ndim))
    qtbot.addWidget(view)
    assert view.dims.last_used == 0
    view.dims._focus_down()
    view.dims._focus_up()
    assert view.dims.last_used == 0

    view.dims.ndim = 5
    view.dims.last_used = 2
    assert view.dims.last_used == 2
    view.dims._focus_down()
    assert view.dims.last_used == 1
    view.dims._focus_up()
    assert view.dims.last_used == 2
    view.dims._focus_up()
    assert view.dims.last_used == 0
    view.dims._focus_down()
    assert view.dims.last_used == 2


def test_changing_display(qtbot):
    """
    Test changing the displayed property of an axis
    """
    ndim = 4
    view = QtDims(Dims(ndim=ndim))
    qtbot.addWidget(view)

    assert view.nsliders == view.dims.ndim
    assert np.sum(view._displayed_sliders) == view.dims.ndim - 2
    assert np.all(
        [
            s.isVisibleTo(view) == d
            for s, d in zip(
                view.slider_widgets, view._displayed_sliders, strict=False
            )
        ]
    )

    # Check changing displayed removes a slider
    view.dims.ndisplay = 3
    assert view.nsliders == view.dims.ndim
    assert np.sum(view._displayed_sliders) == view.dims.ndim - 3
    assert np.all(
        [
            s.isVisibleTo(view) == d
            for s, d in zip(
                view.slider_widgets, view._displayed_sliders, strict=False
            )
        ]
    )


def test_slider_values(qtbot):
    """
    Test the values of a slider stays matched to the values of the dims point.
    """
    ndim = 4
    view = QtDims(Dims(ndim=ndim))
    qtbot.addWidget(view)

    # Check that values of the dimension slider matches the values of the
    # dims point at initialization
    first_slider = view.slider_widgets[0].slider
    assert first_slider.value() == view.dims.point[0]

    # Check that values of the dimension slider matches the values of the
    # dims point after the point has been moved within the dims
    view.dims.set_point(0, 2)
    assert first_slider.value() == view.dims.point[0]

    # Check that values of the dimension slider matches the values of the
    # dims point after the point has been moved within the slider
    first_slider.setValue(1)
    assert first_slider.value() == view.dims.point[0]


def test_slider_range(qtbot):
    """
    Tests range of the slider is matched to the range of the dims
    """
    ndim = 4
    view = QtDims(Dims(ndim=ndim))
    qtbot.addWidget(view)

    # Check the maximum allowed value of the slider is one less
    # than the allowed nsteps of the dims at initialization
    first_slider = view.slider_widgets[0].slider
    assert first_slider.minimum() == 0
    assert first_slider.maximum() == view.dims.nsteps[0] - 1
    assert first_slider.singleStep() == 1

    # Check the maximum allowed value of the slider stays one less
    # than the allowed nsteps of the dims after updates
    view.dims.set_range(0, (1, 5, 2))
    assert first_slider.minimum() == 0
    assert first_slider.maximum() == view.dims.nsteps[0] - 1
    assert first_slider.singleStep() == 1


def test_singleton_dims(qtbot):
    """
    Test singleton dims causes no slider.
    """
    ndim = 4
    dims = Dims(ndim=ndim)
    dims.set_range(0, (0, 0, 1))
    view = QtDims(dims)
    qtbot.addWidget(view)

    # Check that the dims model has been appended to the dims view
    assert view.dims == dims

    # Check the number of displayed sliders is only one
    assert view.nsliders == 4
    assert np.sum(view._displayed_sliders) == 1
    assert np.all(
        [
            s.isVisibleTo(view) == d
            for s, d in zip(
                view.slider_widgets, view._displayed_sliders, strict=False
            )
        ]
    )

    # Change ndisplay to three
    view.dims.ndisplay = 3

    # Check no sliders now shown
    assert np.sum(view._displayed_sliders) == 0

    # Change ndisplay back to two
    view.dims.ndisplay = 2

    # Check only slider now shown
    assert np.sum(view._displayed_sliders) == 1


def test_order_when_changing_ndim(qtbot):
    """
    Test order of the sliders when changing the number of dimensions.
    """
    ndim = 4
    view = QtDims(Dims(ndim=ndim))
    qtbot.addWidget(view)

    # Check that values of the dimension slider matches the values of the
    # dims point after the point has been moved within the dims
    view.dims.set_point(0, 2)
    view.dims.set_point(1, 1)
    for i in range(view.dims.ndim - 2):
        slider = view.slider_widgets[i].slider
        assert slider.value() == view.dims.point[i]

    # Check the matching dimensions and sliders are preserved when
    # dimensions are added
    view.dims.ndim = 5
    for i in range(view.dims.ndim - 2):
        slider = view.slider_widgets[i].slider
        assert slider.value() == view.dims.point[i]

    # Check the matching dimensions and sliders are preserved when dims
    # dimensions are removed
    view.dims.ndim = 4
    for i in range(view.dims.ndim - 2):
        slider = view.slider_widgets[i].slider
        assert slider.value() == view.dims.point[i]

    # Check the matching dimensions and sliders are preserved when dims
    # dimensions are removed
    view.dims.ndim = 3
    for i in range(view.dims.ndim - 2):
        slider = view.slider_widgets[i].slider
        assert slider.value() == view.dims.point[i]


def test_update_dims_labels(qtbot):
    """
    Test that the slider_widget axis labels are updated with the dims model
    and vice versa with eliding capabilites.
    """
    ndim = 4
    view = QtDims(Dims(ndim=ndim))
    qtbot.addWidget(view)

    # set initial widget width and show it to be able to trigger `resizeEvent`
    view.setFixedWidth(100)
    view.show()

    view.dims.axis_labels = list('TZYX')
    assert [w.axis_label.text() for w in view.slider_widgets] == list('TZYX')

    observed_axis_labels_event = False

    def on_axis_labels_changed():
        nonlocal observed_axis_labels_event
        observed_axis_labels_event = True

    view.dims.events.axis_labels.connect(on_axis_labels_changed)
    first_label = view.slider_widgets[0].axis_label
    assert first_label.text() == view.dims.axis_labels[0]

    # check that the label text corresponds with the dims model
    # while being elided on the GUI
    first_label.setText('napari')
    assert first_label.text() == view.dims.axis_labels[0]
    assert '…' in first_label._elidedText()
    assert observed_axis_labels_event

    # increase width to check the full text is shown
    view.setFixedWidth(250)
    assert first_label.text() == view.dims.axis_labels[0]
    assert first_label._elidedText() == view.dims.axis_labels[0]


def test_slider_press_updates_last_used(qtbot):
    """pressing on the slider should update the dims.last_used property"""
    ndim = 5
    view = QtDims(Dims(ndim=ndim))
    qtbot.addWidget(view)

    for i, widg in enumerate(view.slider_widgets):
        widg.slider.sliderPressed.emit()
        if i in [0, 1, 2]:
            # only the first three dims should have visible sliders
            assert widg.isVisibleTo(view)
            assert view.dims.last_used == i
        else:
            # sliders should not be visible for the following dims and the
            # last_used should fallback to the first available dim with a
            # visible slider (dim 0)
            assert not widg.isVisibleTo(view)
            assert view.dims.last_used == 0


@pytest.mark.skipif(
    os.environ.get('CI') and platform == 'win32',
    reason='not working in windows VM',
)
def test_play_button(qtbot):
    """test that the play button and its popup dialog work"""
    ndim = 3
    view = QtDims(Dims(ndim=ndim))
    qtbot.addWidget(view)
    slider = view.slider_widgets[0]
    button = slider.play_button

    # Need looping playback so that it does not stop before we can assert that.
    assert slider.loop_mode == 'loop'
    assert not view.is_playing

    qtbot.mouseClick(button, Qt.LeftButton)
    qtbot.waitUntil(lambda: view.is_playing)

    qtbot.mouseClick(button, Qt.LeftButton)
    qtbot.waitUntil(lambda: not view.is_playing)

    with patch.object(button.popup, 'show_above_mouse') as mock_popup:
        qtbot.mouseClick(button, Qt.RightButton)
        mock_popup.assert_called_once()

    # Check popup updates widget properties (fps, play mode and loop mode)
    button.fpsspin.clear()
    qtbot.keyClicks(button.fpsspin, '11')
    qtbot.keyClick(button.fpsspin, Qt.Key_Enter)
    assert slider.fps == button.fpsspin.value() == 11
    button.reverse_check.setChecked(True)
    assert slider.fps == -button.fpsspin.value() == -11
    button.mode_combo.setCurrentText('once')
    assert slider.loop_mode == button.mode_combo.currentText() == 'once'
    qtbot.waitUntil(view._animation_thread.isFinished)



"""
Custom Qt widgets that serve as native objects that the public-facing elements
wrap.
"""

import contextlib
import inspect
import os
import sys
import time
import warnings
from collections.abc import Mapping, MutableMapping, Sequence
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    Optional,
    Union,
    cast,
)
from weakref import WeakValueDictionary

import numpy as np
from qtpy.QtCore import (
    QEvent,
    QEventLoop,
    QPoint,
    QProcess,
    QRect,
    QSize,
    Qt,
    Slot,
)
from qtpy.QtGui import QHideEvent, QIcon, QShowEvent
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QDockWidget,
    QHBoxLayout,
    QMainWindow,
    QMenu,
    QShortcut,
    QToolTip,
    QWidget,
)

from napari._app_model.constants import MenuId
from napari._app_model.context import create_context, get_context
from napari._qt._qapp_model import build_qmodel_menu
from napari._qt._qapp_model.qactions import add_dummy_actions, init_qactions
from napari._qt._qapp_model.qactions._debug import _is_set_trace_active
from napari._qt._qplugins import (
    _rebuild_npe1_plugins_menu,
    _rebuild_npe1_samples_menu,
)
from napari._qt.dialogs.confirm_close_dialog import ConfirmCloseDialog
from napari._qt.dialogs.preferences_dialog import PreferencesDialog
from napari._qt.dialogs.qt_activity_dialog import QtActivityDialog
from napari._qt.dialogs.qt_notification import NapariQtNotification
from napari._qt.dialogs.shimmed_plugin_dialog import ShimmedPluginDialog
from napari._qt.qt_event_loop import (
    NAPARI_ICON_PATH,
    get_qapp,
    quit_app as quit_app_,
)
from napari._qt.qt_resources import get_stylesheet
from napari._qt.qt_viewer import QtViewer
from napari._qt.threads.status_checker import StatusChecker
from napari._qt.utils import QImg2array, qbytearray_to_str, str_to_qbytearray
from napari._qt.widgets.qt_command_palette import QCommandPalette
from napari._qt.widgets.qt_viewer_dock_widget import (
    _SHORTCUT_DEPRECATION_STRING,
    QtViewerDockWidget,
)
from napari._qt.widgets.qt_viewer_status_bar import ViewerStatusBar
from napari.plugins import (
    menu_item_template as plugin_menu_item_template,
    plugin_manager,
)
from napari.plugins._npe2 import index_npe1_adapters
from napari.settings import get_settings
from napari.utils import perf
from napari.utils._proxies import MappingProxy, PublicOnlyProxy
from napari.utils.events import Event
from napari.utils.geometry import get_center_bbox
from napari.utils.io import imsave
from napari.utils.misc import (
    in_ipython,
    in_jupyter,
    in_python_repl,
    running_as_constructor_app,
)
from napari.utils.notifications import Notification
from napari.utils.theme import _themes, get_system_theme
from napari.utils.translations import trans

if TYPE_CHECKING:
    from magicgui.widgets import Widget
    from qtpy.QtGui import QImage

    from napari.viewer import Viewer

_sentinel = object()


MenuStr = Literal[
    'file_menu',
    'view_menu',
    'layers_menu',
    'plugins_menu',
    'window_menu',
    'help_menu',
]


class _QtMainWindow(QMainWindow):
    # This was added so that someone can patch
    # `napari._qt.qt_main_window._QtMainWindow._window_icon`
    # to their desired window icon
    _window_icon = NAPARI_ICON_PATH

    # To track window instances and facilitate getting the "active" viewer...
    # We use this instead of QApplication.activeWindow for compatibility with
    # IPython usage. When you activate IPython, it will appear that there are
    # *no* active windows, so we want to track the most recently active windows
    _instances: ClassVar[list['_QtMainWindow']] = []

    # `window` is passed through on construction, so it's available to a window
    # provider for dependency injection
    # See https://github.com/napari/napari/pull/4826
    def __init__(
        self, viewer: 'Viewer', window: 'Window', parent=None
    ) -> None:
        super().__init__(parent)
        self._ev = None
        self._window = window
        self._qt_viewer = QtViewer(viewer, show_welcome_screen=True)
        self._quit_app = False

        self.setWindowIcon(QIcon(self._window_icon))
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        center = QWidget(self)
        center.setLayout(QHBoxLayout())
        center.layout().addWidget(self._qt_viewer)
        center.layout().setContentsMargins(4, 0, 4, 0)
        self.setCentralWidget(center)

        self.setWindowTitle(self._qt_viewer.viewer.title)

        self._maximized_flag = False
        self._normal_geometry = QRect()
        self._window_size = None
        self._window_pos = None
        self._old_size = None
        self._positions = []
        self._toggle_menubar_visibility = False

        self._is_close_dialog = {False: True, True: True}
        # this ia sa workaround for #5335 issue. The dict is used to not
        # collide shortcuts for close and close all windows

        act_dlg = QtActivityDialog(self._qt_viewer._welcome_widget)
        self._qt_viewer._welcome_widget.resized.connect(
            act_dlg.move_to_bottom_right
        )
        act_dlg.hide()
        self._activity_dialog = act_dlg

        self.setStatusBar(ViewerStatusBar(self))

        # Prevent QLineEdit based widgets to keep focus even when clicks are
        # done outside the widget. See #1571
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        # Ideally this would be in `NapariApplication` but that is outside of Qt
        self._viewer_context = create_context(self)
        self._viewer_context['is_set_trace_active'] = _is_set_trace_active

        settings = get_settings()

        # TODO:
        # settings.plugins.defaults.call_order = plugin_manager.call_order()

        # set the values in plugins to match the ones saved in settings
        if settings.plugins.call_order is not None:
            plugin_manager.set_call_order(settings.plugins.call_order)

        _QtMainWindow._instances.append(self)

        # since we initialize canvas before the window,
        # we need to manually connect them again.
        handle = self.windowHandle()
        if handle is not None:
            handle.screenChanged.connect(self._qt_viewer.canvas.screen_changed)

        # this is the line that initializes any Qt-based app-model Actions that
        # were defined somewhere in the `_qt` module and imported in init_qactions
        init_qactions()

        with contextlib.suppress(IndexError):
            viewer.cursor.events.position.disconnect(
                viewer.update_status_from_cursor
            )

        self.status_thread = StatusChecker(viewer, parent=self)
        self.status_thread.status_and_tooltip_changed.connect(
            self.set_status_and_tooltip
        )
        viewer.cursor.events.position.connect(
            self.status_thread.trigger_status_update
        )
        settings.appearance.events.update_status_based_on_layer.connect(
            self._toggle_status_thread
        )

        self._command_palette = QCommandPalette(self)

    def _toggle_status_thread(self, event: Event):
        if event.value:
            self.status_thread.start()
        else:
            self.status_thread.terminate()

    def showEvent(self, event: QShowEvent):
        """Override to handle window state changes."""
        settings = get_settings()
        # if event loop is not running, we don't want to start the thread
        # If event loop is running, the loopLevel will be above 0
        if (
            settings.appearance.update_status_based_on_layer
            and QApplication.instance().thread().loopLevel()
        ):
            self.status_thread.start()
        super().showEvent(event)

    def enterEvent(self, a0):
        # as we call show in Viewer constructor, we need to start the thread
        # when the mouse enters the window
        # as first call of showEvent is before the event loop is running
        if (
            get_settings().appearance.update_status_based_on_layer
            and not self.status_thread.isRunning()
        ):
            self.status_thread.start()
        super().enterEvent(a0)

    def hideEvent(self, event: QHideEvent):
        self.status_thread.terminate()
        super().hideEvent(event)

    def set_status_and_tooltip(
        self, status_and_tooltip: tuple[str | dict, str] | None
    ):
        if status_and_tooltip is None:
            return
        self._qt_viewer.viewer.status = status_and_tooltip[0]
        self._qt_viewer.viewer.tooltip.text = status_and_tooltip[1]
        if (
            active := self._qt_viewer.viewer.layers.selection.active
        ) is not None:
            self._qt_viewer.viewer.help = active.help

    def statusBar(self) -> 'ViewerStatusBar':
        return super().statusBar()

    @classmethod
    def current(cls) -> Optional['_QtMainWindow']:
        return cls._instances[-1] if cls._instances else None

    @classmethod
    def current_viewer(cls):
        window = cls.current()
        return window._qt_viewer.viewer if window else None

    def event(self, e: QEvent) -> bool:
        if (
            e.type() == QEvent.Type.ToolTip
            and self._qt_viewer.viewer.tooltip.visible
        ):
            # globalPos is for Qt5 e.globalPosition().toPoint() is for QT6
            # https://doc-snapshots.qt.io/qt6-dev/qmouseevent-obsolete.html#globalPos
            pnt = (
                e.globalPosition().toPoint()
                if hasattr(e, 'globalPosition')
                else e.globalPos()
            )
            QToolTip.showText(pnt, self._qt_viewer.viewer.tooltip.text, self)
        if e.type() in {QEvent.Type.WindowActivate, QEvent.Type.ZOrderChange}:
            # upon activation or raise_, put window at the end of _instances
            with contextlib.suppress(ValueError):
                inst = _QtMainWindow._instances
                inst.append(inst.pop(inst.index(self)))

        res = super().event(e)

        if e.type() == QEvent.Type.Close and e.isAccepted():
            # when we close the MainWindow, remove it from the instance list
            with contextlib.suppress(ValueError):
                _QtMainWindow._instances.remove(self)

        return res

    def showFullScreen(self):
        super().showFullScreen()
        # Handle OpenGL based windows fullscreen issue on Windows.
        # For more info see:
        #  * https://doc.qt.io/qt-6/windows-issues.html#fullscreen-opengl-based-windows
        #  * https://bugreports.qt.io/browse/QTBUG-41309
        #  * https://bugreports.qt.io/browse/QTBUG-104511
        if os.name != 'nt':
            return
        import win32con
        import win32gui

        if self.windowHandle():
            handle = int(self.windowHandle().winId())
            win32gui.SetWindowLong(
                handle,
                win32con.GWL_STYLE,
                win32gui.GetWindowLong(handle, win32con.GWL_STYLE)
                | win32con.WS_BORDER,
            )

    def eventFilter(self, source, event):
        # Handle showing hidden menubar on mouse move event.
        # We do not hide menubar when a menu is being shown or
        # we are not in menubar toggled state
        if (
            QApplication.activePopupWidget() is None
            and hasattr(self, '_toggle_menubar_visibility')
            and self._toggle_menubar_visibility
        ):
            if event.type() == QEvent.Type.MouseMove:
                if self.menuBar().isHidden():
                    rect = self.geometry()
                    # set mouse-sensitive zone to trigger showing the menubar
                    rect.setHeight(25)
                    if rect.contains(event.globalPos()):
                        self.menuBar().show()
                else:
                    rect = QRect(
                        self.menuBar().mapToGlobal(QPoint(0, 0)),
                        self.menuBar().size(),
                    )
                    if not rect.contains(event.globalPos()):
                        self.menuBar().hide()
            elif event.type() == QEvent.Type.Leave and source is self:
                self.menuBar().hide()
        return super().eventFilter(source, event)

    def _load_window_settings(self):
        """
        Load window layout settings from configuration.
        """
        settings = get_settings()
        window_position = settings.application.window_position

        # It's necessary to verify if the window/position value is valid with
        # the current screen.
        if not window_position:
            window_position = (self.x(), self.y())
        else:
            origin_x, origin_y = window_position
            screen = QApplication.screenAt(QPoint(origin_x, origin_y))
            screen_geo = screen.geometry() if screen else None
            if not screen_geo:
                window_position = (self.x(), self.y())

        return (
            settings.application.window_state,
            settings.application.window_size,
            window_position,
            settings.application.window_maximized,
            settings.application.window_fullscreen,
        )

    def _get_window_settings(self):
        """Return current window settings.

        Symmetric to the 'set_window_settings' setter.
        """

        window_fullscreen = self.isFullScreen()
        if window_fullscreen:
            window_maximized = self._maximized_flag
        else:
            window_maximized = self.isMaximized()

        window_state = qbytearray_to_str(self.saveState())
        return (
            window_state,
            self._window_size or (self.width(), self.height()),
            self._window_pos or (self.x(), self.y()),
            window_maximized,
            window_fullscreen,
        )

    def _set_window_settings(
        self,
        window_state,
        window_size,
        window_position,
        window_maximized,
        window_fullscreen,
    ):
        """
        Set window settings.

        Symmetric to the 'get_window_settings' accessor.
        """
        self.setUpdatesEnabled(False)
        self.setWindowState(Qt.WindowState.WindowNoState)

        if window_position:
            window_position = QPoint(*window_position)
            self.move(window_position)

        if window_size:
            window_size = QSize(*window_size)
            self.resize(window_size)

        if window_state:
            self.restoreState(str_to_qbytearray(window_state))

        # Toggling the console visibility is disabled when it is not
        # available, so ensure that it is hidden.
        if in_ipython() or in_jupyter() or in_python_repl():
            self._qt_viewer.dockConsole.setVisible(False)

        if window_fullscreen:
            self._maximized_flag = window_maximized
            self.showFullScreen()
        elif window_maximized:
            self.setWindowState(Qt.WindowState.WindowMaximized)

        self.setUpdatesEnabled(True)

    def _save_current_window_settings(self):
        """Save the current geometry of the main window."""
        (
            window_state,
            window_size,
            window_position,
            window_maximized,
            window_fullscreen,
        ) = self._get_window_settings()

        settings = get_settings()
        if settings.application.save_window_geometry:
            settings.application.window_maximized = window_maximized
            settings.application.window_fullscreen = window_fullscreen
            settings.application.window_position = window_position
            settings.application.window_size = window_size
            settings.application.window_statusbar = (
                not self.statusBar().isHidden()
            )

        if settings.application.save_window_state:
            settings.application.window_state = window_state

    def _warn_on_shimmed_plugins(self) -> None:
        """Warn about shimmed plugins if needed.

        In 0.6.0, a plugin using the deprecated plugin engine will be automatically
        converted so it can be used with npe2. By default, a dialog is displayed
        with each startup listing all shimmed plugins. The user can change this setting
        to only be warned about newly installed shimmed plugins.

        """
        from npe2 import plugin_manager as pm

        settings = get_settings()
        shimmed_plugins = set(pm.get_shimmed_plugins())
        if settings.plugins.only_new_shimmed_plugins_warning:
            new_plugins = (
                shimmed_plugins
                - settings.plugins.already_warned_shimmed_plugins
            )
        else:
            new_plugins = shimmed_plugins

        if new_plugins:
            dialog = ShimmedPluginDialog(self, new_plugins)
            dialog.exec_()

    def close(self, quit_app=False, confirm_need=False):
        """Override to handle closing app or just the window."""
        if not quit_app and not self._qt_viewer.viewer.layers:
            return super().close()
        confirm_need_local = confirm_need and self._is_close_dialog[quit_app]
        self._is_close_dialog[quit_app] = False
        # here we save information that we could request confirmation on close
        # So fi function `close` is called again, we don't ask again but just close
        if (
            not confirm_need_local
            or not get_settings().application.confirm_close_window
            or ConfirmCloseDialog(self, quit_app).exec_()
            == QDialog.DialogCode.Accepted
        ):
            self._quit_app = quit_app
            self._is_close_dialog[quit_app] = True
            # here we inform that confirmation dialog is not open
            self._qt_viewer.dims.stop()
            return super().close()
        self._is_close_dialog[quit_app] = True
        return None
        # here we inform that confirmation dialog is not open

    def close_window(self):
        """Close active dialog or active window."""
        parent = QApplication.focusWidget()
        while parent is not None:
            if isinstance(parent, QMainWindow):
                self.close()
                break

            if isinstance(parent, QDialog):
                parent.close()
                break

            try:
                parent = parent.parent()
            except AttributeError:
                parent = getattr(parent, '_parent', None)

    def show(self, block=False):
        super().show()
        self._qt_viewer.setFocus()
        if block:
            self._ev = QEventLoop()
            self._ev.exec()

    def changeEvent(self, event):
        """Handle window state changes."""
        if event.type() == QEvent.Type.WindowStateChange:
            # TODO: handle maximization issue. When double clicking on the
            # title bar on Mac the resizeEvent is called an varying amount
            # of times which makes it hard to track the original size before
            # maximization.
            condition = (
                self.isMaximized() if os.name == 'nt' else self.isFullScreen()
            )
            if condition and self._old_size is not None:
                if self._positions and len(self._positions) > 1:
                    self._window_pos = self._positions[-2]

                self._window_size = (
                    self._old_size.width(),
                    self._old_size.height(),
                )
            else:
                self._old_size = None
                self._window_pos = None
                self._window_size = None
                self._positions = []

        super().changeEvent(event)

    def keyPressEvent(self, event):
        """Called whenever a key is pressed.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self._qt_viewer.canvas._scene_canvas._backend._keyEvent(
            self._qt_viewer.canvas._scene_canvas.events.key_press, event
        )
        event.accept()

    def keyReleaseEvent(self, event):
        """Called whenever a key is released.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self._qt_viewer.canvas._scene_canvas._backend._keyEvent(
            self._qt_viewer.canvas._scene_canvas.events.key_release, event
        )
        event.accept()

    def resizeEvent(self, event):
        """Override to handle original size before maximizing."""
        # the first resize event will have nonsense positions that we don't
        # want to store (and potentially restore)
        if event.oldSize().isValid():
            self._old_size = event.oldSize()
            self._positions.append((self.x(), self.y()))

            if self._positions and len(self._positions) >= 2:
                self._window_pos = self._positions[-2]
                self._positions = self._positions[-2:]

        super().resizeEvent(event)

    def closeEvent(self, event):
        """This method will be called when the main window is closing.

        Regardless of whether cmd Q, cmd W, or the close button is used...
        """
        if (
            event.spontaneous()
            and get_settings().application.confirm_close_window
            and self._qt_viewer.viewer.layers
            and ConfirmCloseDialog(self, False).exec_() != QDialog.Accepted
        ):
            event.ignore()
            return

        self.status_thread.close_terminate()
        self.status_thread.wait()

        if self._ev and self._ev.isRunning():
            self._ev.quit()

        # Close any floating dockwidgets
        for dock in self.findChildren(QtViewerDockWidget):
            if isinstance(dock, QWidget) and dock.isFloating():
                dock.setFloating(False)

        self._save_current_window_settings()

        # On some versions of Darwin, exiting while fullscreen seems to tickle
        # some bug deep in NSWindow.  This forces the fullscreen keybinding
        # test to complete its draw cycle, then pop back out of fullscreen.
        if self.isFullScreen():
            self.showNormal()
            for _ in range(5):
                time.sleep(0.1)
                QApplication.processEvents()

        self._qt_viewer.dims.stop()

        if self._quit_app:
            quit_app_()

        event.accept()

    def restart(self):
        """Restart the napari application in a detached process."""
        process = QProcess()
        process.setProgram(sys.executable)

        if not running_as_constructor_app():
            process.setArguments(sys.argv)

        process.startDetached()
        self.close(quit_app=True)

    def toggle_menubar_visibility(self):
        """
        Change menubar to be shown or to be hidden and shown on mouse movement.

        For the mouse movement functionality see the `eventFilter` implementation.
        """
        self._toggle_menubar_visibility = not self._toggle_menubar_visibility
        self.menuBar().setVisible(not self._toggle_menubar_visibility)
        return self._toggle_menubar_visibility

    @staticmethod
    @Slot(Notification)
    def show_notification(notification: Notification):
        """Show notification coming from a thread."""
        NapariQtNotification.show_notification(notification)


class Window:
    """Application window that contains the menu bar and viewer.

    Parameters
    ----------
    viewer : napari.components.ViewerModel
        Contained viewer widget.

    Attributes
    ----------
    file_menu : qtpy.QtWidgets.QMenu
        File menu.
    help_menu : qtpy.QtWidgets.QMenu
        Help menu.
    main_menu : qtpy.QtWidgets.QMainWindow.menuBar
        Main menubar.
    view_menu : qtpy.QtWidgets.QMenu
        View menu.
    window_menu : qtpy.QtWidgets.QMenu
        Window menu.
    """

    def __init__(self, viewer: 'Viewer', *, show: bool = True) -> None:
        # create QApplication if it doesn't already exist
        qapp = get_qapp()

        # Dictionary holding dock widgets
        self._wrapped_dock_widgets: MutableMapping[str, QtViewerDockWidget] = (
            WeakValueDictionary()
        )
        self._unnamed_dockwidget_count = 1

        self._pref_dialog = None

        # Connect the Viewer and create the Main Window
        self._qt_window = _QtMainWindow(viewer, self)
        qapp.installEventFilter(self._qt_window)

        # connect theme events before collecting plugin-provided themes
        # to ensure icons from the plugins are generated correctly.
        _themes.events.added.connect(self._add_theme)
        _themes.events.removed.connect(self._remove_theme)

        # discover any themes provided by plugins
        plugin_manager.discover_themes()
        self._setup_existing_themes()

        # import and index all discovered shimmed npe1 plugins
        index_npe1_adapters()

        self._add_menus()
        # TODO: the dummy actions should **not** live on the layerlist context
        # as they are unrelated. However, we do not currently have a suitable
        # enclosing context where we could store these keys, such that they
        # **and** the layerlist context key are available when we update
        # menus. We need a single context to contain all keys required for
        # menu update, so we add them to the layerlist context for now.
        add_dummy_actions(self._qt_viewer.viewer.layers._ctx)
        self._update_theme()
        self._update_theme_font_size()
        get_settings().appearance.events.theme.connect(self._update_theme)
        get_settings().appearance.events.font_size.connect(
            self._update_theme_font_size
        )

        self._add_viewer_dock_widget(self._qt_viewer.dockConsole, tabify=False)
        self._add_viewer_dock_widget(
            self._qt_viewer.dockLayerControls,
            tabify=False,
        )
        self._add_viewer_dock_widget(
            self._qt_viewer.dockLayerList, tabify=False
        )
        if perf.perf_config is not None:
            self._add_viewer_dock_widget(
                self._qt_viewer.dockPerformance, menu=self.window_menu
            )

        viewer.events.help.connect(self._help_changed)
        viewer.events.title.connect(self._title_changed)
        viewer.events.theme.connect(self._update_theme)
        viewer.events.status.connect(self._status_changed)

        if show:
            self.show()
            # Ensure the controls dock uses the minimum height
            self._qt_window.resizeDocks(
                [
                    self._qt_viewer.dockLayerControls,
                    self._qt_viewer.dockLayerList,
                ],
                [self._qt_viewer.dockLayerControls.minimumHeight(), 10000],
                Qt.Orientation.Vertical,
            )
            # TODO: where to put this?
            self._qt_window._warn_on_shimmed_plugins()

    def _setup_existing_themes(self, connect: bool = True):
        """This function is only executed once at the startup of napari
        to connect events to themes that have not been connected yet.

        Parameters
        ----------
        connect : bool
            Determines whether the `connect` or `disconnect` method should be used.
        """
        for theme in _themes.values():
            if connect:
                self._connect_theme(theme)
            else:
                self._disconnect_theme(theme)

    def _connect_theme(self, theme):
        # connect events to update theme. Here, we don't want to pass the event
        # since it won't have the right `value` attribute.
        theme.events.background.connect(self._update_theme_no_event)
        theme.events.foreground.connect(self._update_theme_no_event)
        theme.events.primary.connect(self._update_theme_no_event)
        theme.events.secondary.connect(self._update_theme_no_event)
        theme.events.highlight.connect(self._update_theme_no_event)
        theme.events.text.connect(self._update_theme_no_event)
        theme.events.warning.connect(self._update_theme_no_event)
        theme.events.current.connect(self._update_theme_no_event)
        theme.events.icon.connect(self._update_theme_no_event)
        theme.events.font_size.connect(self._update_theme_no_event)
        theme.events.canvas.connect(
            lambda _: self._qt_viewer.canvas._set_theme_change(
                get_settings().appearance.theme
            )
        )
        # connect console-specific attributes only if QtConsole
        # is present. The `console` is called which might slow
        # things down a little.
        if self._qt_viewer._console:
            theme.events.console.connect(self._qt_viewer.console._update_theme)
            theme.events.syntax_style.connect(
                self._qt_viewer.console._update_theme
            )

    def _disconnect_theme(self, theme):
        theme.events.background.disconnect(self._update_theme_no_event)
        theme.events.foreground.disconnect(self._update_theme_no_event)
        theme.events.primary.disconnect(self._update_theme_no_event)
        theme.events.secondary.disconnect(self._update_theme_no_event)
        theme.events.highlight.disconnect(self._update_theme_no_event)
        theme.events.text.disconnect(self._update_theme_no_event)
        theme.events.warning.disconnect(self._update_theme_no_event)
        theme.events.current.disconnect(self._update_theme_no_event)
        theme.events.icon.disconnect(self._update_theme_no_event)
        theme.events.font_size.disconnect(self._update_theme_no_event)
        theme.events.canvas.disconnect(
            lambda _: self._qt_viewer.canvas._set_theme_change(
                get_settings().appearance.theme
            )
        )
        # disconnect console-specific attributes only if QtConsole
        # is present and they were previously connected
        if self._qt_viewer._console:
            theme.events.console.disconnect(
                self._qt_viewer.console._update_theme
            )
            theme.events.syntax_style.disconnect(
                self._qt_viewer.console._update_theme
            )

    def _add_theme(self, event):
        """Add new theme and connect events."""
        theme = event.value
        self._connect_theme(theme)

    def _remove_theme(self, event):
        """Remove theme and disconnect events."""
        theme = event.value
        self._disconnect_theme(theme)

    @property
    def qt_viewer(self):
        warnings.warn(
            trans._(
                'Public access to Window.qt_viewer is deprecated and will be removed in\n'
                'v0.7.0. It is considered an "implementation detail" of the napari\napplication, '
                'not part of the napari viewer model. If your use case\n'
                'requires access to qt_viewer, please open an issue to discuss.',
                deferred=True,
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self._qt_window._qt_viewer

    @property
    def _qt_viewer(self):
        # this is starting to be "vestigial"... this property could be removed
        return self._qt_window._qt_viewer

    @property
    def _status_bar(self):
        # TODO: remove from window
        return self._qt_window.statusBar()

    def _update_menu_state(self, menu: MenuStr):
        """Update enabled/visible state of menu item with context."""
        layerlist = self._qt_viewer.viewer.layers
        menu_model = getattr(self, menu)
        menu_model.update_from_context(get_context(layerlist))

    def _update_file_menu_state(self):
        self._update_menu_state('file_menu')

    def _update_view_menu_state(self):
        self._update_menu_state('view_menu')

    def _update_layers_menu_state(self):
        self._update_menu_state('layers_menu')

    def _update_window_menu_state(self):
        self._update_menu_state('window_menu')

    def _update_plugins_menu_state(self):
        self._update_menu_state('plugins_menu')

    def _update_help_menu_state(self):
        self._update_menu_state('help_menu')

    def _update_debug_menu_state(self):
        viewer_ctx = get_context(self._qt_window)
        self._debug_menu.update_from_context(viewer_ctx)

    # TODO: Remove once npe1 deprecated
    def _setup_npe1_samples_menu(self):
        """Register npe1 sample data, build menu and connect to events."""
        plugin_manager.discover_sample_data()
        plugin_manager.events.enabled.connect(_rebuild_npe1_samples_menu)
        plugin_manager.events.disabled.connect(_rebuild_npe1_samples_menu)
        plugin_manager.events.registered.connect(_rebuild_npe1_samples_menu)
        plugin_manager.events.unregistered.connect(_rebuild_npe1_samples_menu)
        _rebuild_npe1_samples_menu()

    # TODO: Remove once npe1 deprecated
    def _setup_npe1_plugins_menu(self):
        """Register npe1 widgets, build menu and connect to events"""
        plugin_manager.discover_widgets()
        plugin_manager.events.registered.connect(_rebuild_npe1_plugins_menu)
        plugin_manager.events.disabled.connect(_rebuild_npe1_plugins_menu)
        plugin_manager.events.unregistered.connect(_rebuild_npe1_plugins_menu)
        _rebuild_npe1_plugins_menu()

    def _handle_trace_file_on_start(self):
        """Start trace of `trace_file_on_start` config set."""
        from napari._qt._qapp_model.qactions._debug import _start_trace

        if perf.perf_config:
            path = perf.perf_config.trace_file_on_start
            if path is not None:
                # Config option "trace_file_on_start" means immediately
                # start tracing to that file. This is very useful if you
                # want to create a trace every time you start napari,
                # without having to start it from the debug menu.
                _start_trace(path)

    def _add_menus(self):
        """Add menubar to napari app."""
        # TODO: move this to _QMainWindow... but then all of the Menu()
        # items will not have easy access to the methods on this Window obj.

        self.main_menu = self._qt_window.menuBar()
        # Menubar shortcuts are only active when the menubar is visible.
        # Therefore, we set a global shortcut not associated with the menubar
        # to toggle visibility, *but*, in order to not shadow the menubar
        # shortcut, we disable it, and only enable it when the menubar is
        # hidden. See this stackoverflow link for details:
        # https://stackoverflow.com/questions/50537642/how-to-keep-the-shortcuts-of-a-hidden-widget-in-pyqt5
        self._main_menu_shortcut = QShortcut('Ctrl+M', self._qt_window)
        self._main_menu_shortcut.setEnabled(False)
        self._main_menu_shortcut.activated.connect(
            self._toggle_menubar_visible
        )
        # file menu
        self.file_menu = build_qmodel_menu(
            MenuId.MENUBAR_FILE, title=trans._('&File'), parent=self._qt_window
        )
        self._setup_npe1_samples_menu()
        self.file_menu.aboutToShow.connect(
            self._update_file_menu_state,
        )
        self.main_menu.addMenu(self.file_menu)
        # view menu
        self.view_menu = build_qmodel_menu(
            MenuId.MENUBAR_VIEW, title=trans._('&View'), parent=self._qt_window
        )
        self.view_menu.aboutToShow.connect(
            self._update_view_menu_state,
        )
        self.main_menu.addMenu(self.view_menu)
        # layers menu
        self.layers_menu = build_qmodel_menu(
            MenuId.MENUBAR_LAYERS,
            title=trans._('&Layers'),
            parent=self._qt_window,
        )
        self.layers_menu.aboutToShow.connect(
            self._update_layers_menu_state,
        )
        self.main_menu.addMenu(self.layers_menu)
        # plugins menu
        self.plugins_menu = build_qmodel_menu(
            MenuId.MENUBAR_PLUGINS,
            title=trans._('&Plugins'),
            parent=self._qt_window,
        )
        self._setup_npe1_plugins_menu()
        self.plugins_menu.aboutToShow.connect(
            self._update_plugins_menu_state,
        )
        self.main_menu.addMenu(self.plugins_menu)
        # debug menu (optional)
        if perf.perf_config is not None:
            self._debug_menu = build_qmodel_menu(
                MenuId.MENUBAR_DEBUG,
                title=trans._('&Debug'),
                parent=self._qt_window,
            )
            self._handle_trace_file_on_start()
            self._debug_menu.aboutToShow.connect(
                self._update_debug_menu_state,
            )
            self.main_menu.addMenu(self._debug_menu)
        # window menu
        self.window_menu = build_qmodel_menu(
            MenuId.MENUBAR_WINDOW,
            title=trans._('&Window'),
            parent=self._qt_window,
        )
        self.plugins_menu.aboutToShow.connect(
            self._update_window_menu_state,
        )
        self.main_menu.addMenu(self.window_menu)
        # help menu
        self.help_menu = build_qmodel_menu(
            MenuId.MENUBAR_HELP, title=trans._('&Help'), parent=self._qt_window
        )
        self.help_menu.aboutToShow.connect(
            self._update_help_menu_state,
        )
        self.main_menu.addMenu(self.help_menu)

    def _toggle_menubar_visible(self):
        """Toggle visibility of app menubar.

        This function also disables or enables a global keyboard shortcut to
        show the menubar, since menubar shortcuts are only available while the
        menubar is visible.
        """
        toggle_menubar_visibility = self._qt_window.toggle_menubar_visibility()
        self._main_menu_shortcut.setEnabled(toggle_menubar_visibility)

    def _toggle_command_palette(self):
        """Toggle the visibility of the command palette."""
        palette = self._qt_window._command_palette
        if palette.isVisible():
            palette.hide()
        else:
            palette.update_context(self._qt_window)
            palette.show()

    def _toggle_fullscreen(self):
        """Toggle fullscreen mode."""
        if self._qt_window.isFullScreen():
            self._qt_window.showNormal()
        else:
            self._qt_window.showFullScreen()

    def _toggle_play(self):
        """Toggle play."""
        if self._qt_viewer.dims.is_playing:
            self._qt_viewer.dims.stop()
        else:
            axis = self._qt_viewer.viewer.dims.last_used or 0
            self._qt_viewer.dims.play(axis)

    def add_plugin_dock_widget(
        self,
        plugin_name: str,
        widget_name: str | None = None,
        tabify: bool = False,
    ) -> tuple[QtViewerDockWidget, Any]:
        """Add plugin dock widget if not already added.

        Parameters
        ----------
        plugin_name : str
            Name of a plugin providing a widget
        widget_name : str, optional
            Name of a widget provided by `plugin_name`. If `None`, and the
            specified plugin provides only a single widget, that widget will be
            returned, otherwise a ValueError will be raised, by default None
        tabify : bool
            Flag to tabify dock widget or not.

        Returns
        -------
        tuple
            A 2-tuple containing (the DockWidget instance, the plugin widget
            instance).
        """
        from napari.plugins import _npe2

        widget_class = None
        dock_kwargs = {}

        if result := _npe2.get_widget_contribution(plugin_name, widget_name):
            widget_class, widget_name = result

        if widget_class is None:
            widget_class, dock_kwargs = plugin_manager.get_widget(
                plugin_name, widget_name
            )

        if not widget_name:
            # if widget_name wasn't provided, `get_widget` will have
            # ensured that there is a single widget available.
            widget_name = next(
                iter(plugin_manager._wrapped_dock_widgets[plugin_name])
            )

        full_name = plugin_menu_item_template.format(plugin_name, widget_name)
        if full_name in self._wrapped_dock_widgets:
            dock_widget = self._wrapped_dock_widgets[full_name]
            return dock_widget, dock_widget.inner_widget()

        wdg = _instantiate_dock_widget(
            widget_class, cast('Viewer', self._qt_viewer.viewer)
        )

        # Add dock widget
        dock_kwargs.pop('name', None)
        dock_widget = self.add_dock_widget(
            wdg, name=full_name, tabify=tabify, **dock_kwargs
        )
        return dock_widget, wdg

    def _add_plugin_function_widget(self, plugin_name: str, widget_name: str):
        """Add plugin function widget if not already added.

        Parameters
        ----------
        plugin_name : str
            Name of a plugin providing a widget
        widget_name : str, optional
            Name of a widget provided by `plugin_name`. If `None`, and the
            specified plugin provides only a single widget, that widget will be
            returned, otherwise a ValueError will be raised, by default None
        """
        full_name = plugin_menu_item_template.format(plugin_name, widget_name)
        if full_name in self._wrapped_dock_widgets:
            return None

        func = plugin_manager._function_widgets[plugin_name][widget_name]

        # Add function widget
        return self.add_function_widget(
            func, name=full_name, area=None, allowed_areas=None
        )

    def add_dock_widget(
        self,
        widget: Union[QWidget, 'Widget'],
        *,
        name: str = '',
        area: str | None = None,
        allowed_areas: Sequence[str] | None = None,
        shortcut=_sentinel,
        add_vertical_stretch=True,
        tabify: bool = False,
        menu: QMenu | None = None,
    ):
        """Convenience method to add a QDockWidget to the main window.

        If name is not provided a generic name will be addded to avoid
        `saveState` warnings on close.

        Parameters
        ----------
        widget : QWidget
            `widget` will be added as QDockWidget's main widget.
        name : str, optional
            Name of dock widget to appear in window menu.
        area : str
            Side of the main window to which the new dock widget will be added.
            Must be in {'left', 'right', 'top', 'bottom'}
        allowed_areas : list[str], optional
            Areas, relative to the main window, that the widget is allowed dock.
            Each item in list must be in {'left', 'right', 'top', 'bottom'}
            By default, all areas are allowed.
        shortcut : str, optional
            Keyboard shortcut to appear in dropdown menu.
        add_vertical_stretch : bool, optional
            Whether to add stretch to the bottom of vertical widgets (pushing
            widgets up towards the top of the allotted area, instead of letting
            them distribute across the vertical space).  By default, True.

            .. deprecated:: 0.4.8

                The shortcut parameter is deprecated since version 0.4.8, please use
                the action and shortcut manager APIs. The new action manager and
                shortcut API allow user configuration and localization.
        tabify : bool
            Flag to tabify dock widget or not.
        menu : QMenu, optional
            Menu bar to add toggle action to. If `None` nothing added to menu.

        Returns
        -------
        dock_widget : QtViewerDockWidget
            `dock_widget` that can pass viewer events.
        """
        if not name:
            with contextlib.suppress(AttributeError):
                name = widget.objectName()
            name = name or trans._(
                'Dock widget {number}',
                number=self._unnamed_dockwidget_count,
            )

            self._unnamed_dockwidget_count += 1

        if area is None:
            settings = get_settings()
            area = settings.application.plugin_widget_positions.get(
                name, 'right'
            )

        if shortcut is not _sentinel:
            warnings.warn(
                _SHORTCUT_DEPRECATION_STRING.format(shortcut=shortcut),
                FutureWarning,
                stacklevel=2,
            )
            dock_widget = QtViewerDockWidget(
                self._qt_viewer,
                widget,
                name=name,
                area=area,
                allowed_areas=allowed_areas,
                shortcut=shortcut,
                add_vertical_stretch=add_vertical_stretch,
            )
        else:
            dock_widget = QtViewerDockWidget(
                self._qt_viewer,
                widget,
                name=name,
                area=area,
                allowed_areas=allowed_areas,
                add_vertical_stretch=add_vertical_stretch,
            )

        self._add_viewer_dock_widget(dock_widget, tabify=tabify, menu=menu)

        if hasattr(widget, 'reset_choices'):
            # Keep the dropdown menus in the widget in sync with the layer model
            # if widget has a `reset_choices`, which is true for all magicgui
            # `CategoricalWidget`s
            layers_events = self._qt_viewer.viewer.layers.events
            layers_events.inserted.connect(widget.reset_choices)
            layers_events.removed.connect(widget.reset_choices)
            layers_events.reordered.connect(widget.reset_choices)

        # Add dock widget to dictionary
        self._wrapped_dock_widgets[dock_widget.name] = dock_widget

        return dock_widget

    @property
    def _dock_widgets(self) -> MutableMapping[str, QtViewerDockWidget]:
        """Access `_wrapped_dock_widgets` with warning.

        Before napari 0.6.2, ``_wrapped_dock_widgets`` was just
        ``_dock_widgets``. Even though it was private, many
        resources pointed to its use, as there was no public alternative.
        Now that `dock_widgets` is provided, we want to make sure
        that people stop using the private `_dock_widgets`.
        """
        # As many plugins uses `_dock_widget` to access one widget from the
        # other widget we should keep this name for a longer period
        warnings.warn(
            'The `_dock_widgets` property is private and should not be used in any plugin code. '
            'Please use the `dock_widgets` property instead.',
            FutureWarning,
            stacklevel=2,
        )
        return self._wrapped_dock_widgets

    @property
    def dock_widgets(self) -> Mapping[str, 'QWidget | Widget']:
        """Read only mapping of widgets docked in napari window.

        For wrapping QtViewerDockWidget use `dock_widgets` property.
        """
        return InnerWidgetMappingProxy(self._wrapped_dock_widgets)

    def _add_viewer_dock_widget(
        self,
        dock_widget: QtViewerDockWidget,
        tabify: bool = False,
        menu: QMenu | None = None,
    ):
        """Add a QtViewerDockWidget to the main window

        If other widgets already present in area then will tabify.

        Parameters
        ----------
        dock_widget : QtViewerDockWidget
            `dock_widget` will be added to the main window.
        tabify : bool
            Flag to tabify dockwidget or not.
        menu : QMenu, optional
            Menu bar to add toggle action to. If `None` nothing added to menu.
        """
        # Find if any other dock widgets are currently in area
        current_dws_in_area = [
            dw
            for dw in self._qt_window.findChildren(QDockWidget)
            if self._qt_window.dockWidgetArea(dw) == dock_widget.qt_area
        ]
        self._qt_window.addDockWidget(dock_widget.qt_area, dock_widget)

        # If another dock widget present in area then tabify
        if current_dws_in_area:
            if tabify:
                self._qt_window.tabifyDockWidget(
                    current_dws_in_area[-1], dock_widget
                )
                dock_widget.show()
                dock_widget.raise_()
            elif dock_widget.area in ('right', 'left'):
                _wdg = [*current_dws_in_area, dock_widget]
                # add sizes to push lower widgets up
                sizes = list(range(1, len(_wdg) * 4, 4))
                self._qt_window.resizeDocks(
                    _wdg, sizes, Qt.Orientation.Vertical
                )

        if menu:
            action = dock_widget.toggleViewAction()
            action.setStatusTip(dock_widget.name)
            action.setText(dock_widget.name)
            import warnings

            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FutureWarning)
                # deprecating with 0.4.8, but let's try to keep compatibility.
                shortcut = dock_widget.shortcut
            if shortcut is not None:
                action.setShortcut(shortcut)

            menu.addAction(action)

        # see #3663, to fix #3624 more generally
        dock_widget.setFloating(False)

    def _remove_dock_widget(self, event) -> None:
        names = list(self._wrapped_dock_widgets.keys())
        for widget_name in names:
            if event.value in widget_name:
                # remove this widget
                widget = self._wrapped_dock_widgets[widget_name]
                self.remove_dock_widget(widget)

    def remove_dock_widget(self, widget: QWidget, menu=None):
        """Removes specified dock widget.

        If a QDockWidget is not provided, the existing QDockWidgets will be
        searched for one whose inner widget (``.widget()``) is the provided
        ``widget``.

        Parameters
        ----------
        widget : QWidget | str
            If widget == 'all', all docked widgets will be removed.
        menu : QMenu, optional
            Menu bar to remove toggle action from. If `None` nothing removed
            from menu.
        """
        if widget == 'all':
            for dw in list(self._wrapped_dock_widgets.values()):
                self.remove_dock_widget(dw)
            return

        if not isinstance(widget, QDockWidget):
            dw: QDockWidget
            for dw in self._qt_window.findChildren(QDockWidget):
                if dw.widget() is widget:
                    _dw: QDockWidget = dw
                    break
            else:
                raise LookupError(
                    trans._(
                        'Could not find a dock widget containing: {widget}',
                        deferred=True,
                        widget=widget,
                    )
                )
        else:
            _dw = widget

        if _dw.widget():
            _dw.widget().setParent(None)
        self._qt_window.removeDockWidget(_dw)
        if menu is not None:
            menu.removeAction(_dw.toggleViewAction())

        # Remove dock widget from dictionary
        self._wrapped_dock_widgets.pop(_dw.name, None)

        # Deleting the dock widget means any references to it will no longer
        # work but it's not really useful anyway, since the inner widget has
        # been removed. and anyway: people should be using add_dock_widget
        # rather than directly using _add_viewer_dock_widget
        _dw.deleteLater()

    def add_function_widget(
        self,
        function,
        *,
        magic_kwargs=None,
        name: str = '',
        area=None,
        allowed_areas=None,
        shortcut=_sentinel,
    ):
        """Turn a function into a dock widget via magicgui.

        Parameters
        ----------
        function : callable
            Function that you want to add.
        magic_kwargs : dict, optional
            Keyword arguments to :func:`magicgui.magicgui` that
            can be used to specify widget.
        name : str, optional
            Name of dock widget to appear in window menu.
        area : str, optional
            Side of the main window to which the new dock widget will be added.
            Must be in {'left', 'right', 'top', 'bottom'}. If not provided the
            default will be determined by the widget.layout, with 'vertical'
            layouts appearing on the right, otherwise on the bottom.
        allowed_areas : list[str], optional
            Areas, relative to main window, that the widget is allowed dock.
            Each item in list must be in {'left', 'right', 'top', 'bottom'}
            By default, only provided areas is allowed.
        shortcut : str, optional
            Keyboard shortcut to appear in dropdown menu.

        Returns
        -------
        dock_widget : QtViewerDockWidget
            `dock_widget` that can pass viewer events.
        """
        from magicgui import magicgui

        if magic_kwargs is None:
            magic_kwargs = {
                'auto_call': False,
                'call_button': 'run',
                'layout': 'vertical',
            }

        widget = magicgui(function, **magic_kwargs or {})

        if area is None:
            area = 'right' if str(widget.layout) == 'vertical' else 'bottom'
        if allowed_areas is None:
            allowed_areas = [area]
        if shortcut is not _sentinel:
            return self.add_dock_widget(
                widget,
                name=name or function.__name__.replace('_', ' '),
                area=area,
                allowed_areas=allowed_areas,
                shortcut=shortcut,
            )

        return self.add_dock_widget(
            widget,
            name=name or function.__name__.replace('_', ' '),
            area=area,
            allowed_areas=allowed_areas,
        )

    def resize(self, width, height):
        """Resize the window.

        Parameters
        ----------
        width : int
            Width in logical pixels.
        height : int
            Height in logical pixels.
        """
        self._qt_window.resize(width, height)

    def set_geometry(self, left, top, width, height):
        """Set the geometry of the widget

        Parameters
        ----------
        left : int
            X coordinate of the upper left border.
        top : int
            Y coordinate of the upper left border.
        width : int
            Width of the rectangle shape of the window.
        height : int
            Height of the rectangle shape of the window.
        """
        self._qt_window.setGeometry(left, top, width, height)

    def geometry(self) -> tuple[int, int, int, int]:
        """Get the geometry of the widget

        Returns
        -------
        left : int
            X coordinate of the upper left border.
        top : int
            Y coordinate of the upper left border.
        width : int
            Width of the rectangle shape of the window.
        height : int
            Height of the rectangle shape of the window.
        """
        rect = self._qt_window.geometry()
        return rect.left(), rect.top(), rect.width(), rect.height()

    def show(self, *, block=False):
        """Resize, show, and bring forward the window.

        Raises
        ------
        RuntimeError
            If the viewer.window has already been closed and deleted.
        """
        settings = get_settings()
        try:
            self._qt_window.show(block=block)
        except (AttributeError, RuntimeError) as e:
            raise RuntimeError(
                trans._(
                    'This viewer has already been closed and deleted. Please create a new one.',
                    deferred=True,
                )
            ) from e

        if settings.application.first_time:
            settings.application.first_time = False
            try:
                self._qt_window.resize(self._qt_window.layout().sizeHint())
            except (AttributeError, RuntimeError) as e:
                raise RuntimeError(
                    trans._(
                        'This viewer has already been closed and deleted. Please create a new one.',
                        deferred=True,
                    )
                ) from e
        else:
            try:
                if settings.application.save_window_geometry:
                    self._qt_window._set_window_settings(
                        *self._qt_window._load_window_settings()
                    )
            except Exception as err:  # noqa: BLE001
                import warnings

                warnings.warn(
                    trans._(
                        'The window geometry settings could not be loaded due to the following error: {err}',
                        deferred=True,
                        err=err,
                    ),
                    category=RuntimeWarning,
                    stacklevel=2,
                )

        # Resize axis labels now that window is shown
        self._qt_viewer.dims._resize_axis_labels()

        # We want to bring the viewer to the front when
        # A) it is our own event loop OR we are running in jupyter
        # B) it is not the first time a QMainWindow is being created

        # `app_name` will be "napari" iff the application was instantiated in
        # get_qapp(). isActiveWindow() will be True if it is the second time a
        # _qt_window has been created.
        # See #721, #732, #735, #795, #1594
        app_name = QApplication.instance().applicationName()
        if (
            app_name == 'napari' or in_jupyter()
        ) and self._qt_window.isActiveWindow():
            self.activate()

    def activate(self):
        """Make the viewer the currently active window."""
        self._qt_window.raise_()  # for macOS
        self._qt_window.activateWindow()  # for Windows

    def _update_theme_no_event(self):
        self._update_theme()

    def _update_theme_font_size(self, event=None):
        settings = get_settings()
        font_size = event.value if event else settings.appearance.font_size
        extra_variables = {'font_size': f'{font_size}pt'}
        self._update_theme(extra_variables=extra_variables)

    def _update_theme(self, event=None, extra_variables=None):
        """Update widget color theme."""
        if extra_variables is None:
            extra_variables = {}
        settings = get_settings()
        with contextlib.suppress(AttributeError, RuntimeError):
            value = event.value if event else settings.appearance.theme
            self._qt_viewer.viewer.theme = value
            actual_theme_name = value
            if value == 'system':
                # system isn't a theme, so get the name
                actual_theme_name = get_system_theme()
            # check `font_size` value is always passed when updating style
            if 'font_size' not in extra_variables:
                extra_variables.update(
                    {'font_size': f'{settings.appearance.font_size}pt'}
                )
            # set the style sheet with the theme name and extra_variables
            style_sheet = get_stylesheet(
                actual_theme_name, extra_variables=extra_variables
            )
            self._qt_window.setStyleSheet(style_sheet)
            self._qt_viewer.setStyleSheet(style_sheet)
            if self._qt_viewer._console:
                self._qt_viewer._console._update_theme(style_sheet=style_sheet)

    def _status_changed(self, event):
        """Update status bar.

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.
        """
        if not hasattr(self, '_qt_window'):
            return
        if isinstance(event.value, str):
            self._status_bar.setStatusText(event.value)
        else:
            status_info = event.value
            self._status_bar.setStatusText(
                layer_base=status_info['layer_base'],
                source_type=status_info['source_type'],
                plugin=status_info['plugin'],
                coordinates=status_info['coordinates'],
            )

    def _title_changed(self, event):
        """Update window title.

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.
        """
        if hasattr(self, '_qt_window'):
            self._qt_window.setWindowTitle(event.value)

    def _help_changed(self, event):
        """Update help message on status bar.

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.
        """
        if hasattr(self, '_qt_window'):
            self._status_bar.setHelpText(event.value)

    def _restart(self):
        """Restart the napari application."""
        if hasattr(self, '_qt_window'):
            self._qt_window.restart()

    def _screenshot(
        self,
        size: tuple[int, int] | None = None,
        scale: float | None = None,
        flash: bool = True,
        canvas_only: bool = False,
        fit_to_data_extent: bool = False,
    ) -> 'QImage':
        """Capture screenshot of the currently displayed viewer.

        Parameters
        ----------
        flash : bool
            Flag to indicate whether flash animation should be shown after
            the screenshot was captured.
        size : tuple of two ints, optional
            Size (resolution height x width) of the screenshot. By default, the
            currently displayed size. Only used if `canvas_only` is True. This
            argument is ignored if fit_to_data_extent is set to True.
        scale : float, optional
            Scale factor used to increase resolution of canvas for the screenshot.
            By default, the currently displayed resolution.
            Only used if `canvas_only` is True.
        canvas_only : bool
            If True, screenshot shows only the image display canvas, and
            if False include the napari viewer frame in the screenshot,
            By default, True.
        fit_to_data_extent: bool
            Tightly fit the canvas around the data to prevent margins from
            showing in the screenshot. If False, a screenshot of the currently
            visible canvas will be generated.

        Returns
        -------
        img : QImage
        """
        from napari._qt.utils import add_flash_animation

        canvas = self._qt_viewer.canvas
        prev_size = canvas.size
        camera = self._qt_viewer.viewer.camera
        old_center = camera.center
        old_zoom = camera.zoom
        ndisplay = self._qt_viewer.viewer.dims.ndisplay

        # Part 1: validate incompatible parameters
        if not canvas_only and (
            fit_to_data_extent or size is not None or scale is not None
        ):
            raise ValueError(
                trans._(
                    'scale, size, and fit_to_data_extent can only be set for '
                    'canvas_only screenshots.',
                    deferred=True,
                )
            )
        if size is not None and len(size) != 2:
            raise ValueError(
                trans._(
                    'screenshot size must be 2 values, got {len_size}',
                    deferred=True,
                    len_size=len(size),
                )
            )

        # Part 2: compute canvas size and view based on parameters
        if fit_to_data_extent:
            # Use the same scene parameter calculations as in viewer_model.fit_to_view
            extent, scene_size, _ = (
                self._qt_viewer.viewer._get_scene_parameters()
            )
            extent_scale = min(
                self._qt_viewer.viewer.layers.extent.step[-ndisplay:]
            )

            if ndisplay == 3:
                scene_size = self._qt_viewer.viewer._calculate_bounding_box(
                    extent=extent,
                    view_direction=self._qt_viewer.viewer.camera.view_direction,
                    up_direction=self._qt_viewer.viewer.camera.up_direction,
                )

            # adjust size by the scale, to return the size in real pixels
            size = np.ceil(scene_size / extent_scale).astype(int)

        if size is not None:
            size = np.asarray(size) / self._qt_window.devicePixelRatio()
        else:
            size = np.asarray(prev_size)

        if scale is not None:
            # multiply canvas dimensions by the scale factor to get new size
            size *= scale

        # Part 3: take the screenshot
        if canvas_only:
            if fit_to_data_extent:
                grid_shape = self._qt_viewer.viewer.grid.actual_shape(
                    len(self._qt_viewer.viewer.layers)
                )
                canvas.size = tuple((size * grid_shape).astype(int))
                # tight view around data
                self._qt_viewer.viewer.fit_to_view(margin=0)
            else:
                canvas.size = tuple(size.astype(int))

            try:
                img = canvas.screenshot()
                if flash:
                    add_flash_animation(self._qt_viewer._welcome_widget)
            finally:
                # make sure we always go back to the right canvas size
                canvas.size = prev_size
                camera.center = old_center
                camera.zoom = old_zoom
        else:
            img = self._qt_window.grab().toImage()
            if flash:
                add_flash_animation(self._qt_window)
        return img

    def export_figure(
        self,
        path: str | None = None,
        scale: float = 1,
        flash=True,
    ) -> np.ndarray:
        """Export an image of the full extent of the displayed layer data.

        This function finds a tight boundary around the data, resets the view
        around that boundary (and, when scale=1, such that 1 captured pixel is
        equivalent to one data pixel), takes a screenshot, then restores the
        previous zoom and canvas sizes.

        Parameters
        ----------
        path : str, optional
            Filename for saving screenshot image.
        scale : float
            Scale factor used to increase resolution of canvas for the
            screenshot. By default, a scale of 1.
        flash : bool
            Flag to indicate whether flash animation should be shown after
            the screenshot was captured.
            By default, True.

        Returns
        -------
        image : array
            Numpy array of type ubyte and shape (h, w, 4). Index [0, 0] is the
            upper-left corner of the rendered region.
        """
        if not isinstance(scale, float | int):
            raise TypeError(
                trans._(
                    'Scale must be a float or an int.',
                    deferred=True,
                )
            )
        img = QImg2array(
            self._screenshot(
                scale=scale,
                flash=flash,
                canvas_only=True,
                fit_to_data_extent=True,
            )
        )
        if path is not None:
            imsave(path, img)
        return img

    def export_rois(
        self,
        rois: list[np.ndarray],
        paths: str | Path | list[str | Path] | None = None,
        scale: float | None = None,
    ):
        """Export the given rectangular rois to specified file paths.

        For each shape, moves the camera to the center of the shape
        and adjust the canvas size to fit the shape.
        Note: The shape height and width can be of type float.
        However, the canvas size only accepts a tuple of integers.
        This can result in slight misalignment.

        Parameters
        ----------
        rois: list[np.ndarray]
            A list of arrays  with each being of shape (4, 2) representing
            a rectangular roi.
        paths: str, Path, list[str, Path], optional
            Where to save the rois. If a string or a Path, a directory will
            be created if it does not exist yet and screenshots will be
            saved with filename `roi_{n}.png` where n is the nth roi. If
            paths is a list of either string or paths, these need to be the
            full paths of where to store each individual roi. In this case
            the length of the list and the number of rois must match.
            If None, the screenshots will only be returned and not saved
            to disk.
        scale: float, optional
            Scale factor used to increase resolution of canvas for the screenshot.
            By default, uses the displayed scale.

        Returns
        -------
        screenshot_list: list
            The list with roi screenshots.

        """
        if (
            paths is not None
            and isinstance(paths, list)
            and len(paths) != len(rois)
        ):
            raise ValueError(
                trans._(
                    'The number of file paths does not match the number of ROI shapes',
                    deferred=True,
                )
            )

        if isinstance(paths, str | Path):
            storage_dir = Path(paths).expanduser()
            storage_dir.mkdir(parents=True, exist_ok=True)
            paths = [storage_dir / f'roi_{n}.png' for n in range(len(rois))]

        if self._qt_viewer.viewer.dims.ndisplay > 2:
            raise NotImplementedError(
                "'export_rois' is not implemented for 3D view."
            )

        screenshot_list = []
        camera = self._qt_viewer.viewer.camera
        start_camera_center = camera.center
        start_camera_zoom = camera.zoom
        canvas = self._qt_viewer.canvas
        prev_size = canvas.size

        visible_dims = list(self._qt_viewer.viewer.dims.displayed)
        step = min(self._qt_viewer.viewer.layers.extent.step[visible_dims])

        for index, roi in enumerate(rois):
            center_coord, height, width = get_center_bbox(roi)
            camera.center = center_coord
            canvas.size = (int(height / step), int(width / step))

            camera.zoom = 1 / step
            path = paths[index] if paths is not None else None
            screenshot_list.append(
                self.screenshot(path=path, canvas_only=True, scale=scale)
            )

        canvas.size = prev_size
        camera.center = start_camera_center
        camera.zoom = start_camera_zoom

        return screenshot_list

    def screenshot(
        self, path=None, size=None, scale=None, flash=True, canvas_only=False
    ):
        """Take currently displayed viewer and convert to an image array.

        Parameters
        ----------
        path : str, Path
            Filename for saving screenshot image.
        size : tuple (int, int)
            Size (resolution) of the screenshot. By default, the currently displayed size.
            Only used if `canvas_only` is True.
        scale : float
            Scale factor used to increase resolution of canvas for the screenshot.
            By default, the currently displayed resolution.
            Only used if `canvas_only` is True.
        flash : bool
            Flag to indicate whether flash animation should be shown after
            the screenshot was captured.
        canvas_only : bool
            If True, screenshot shows only the image display canvas, and
            if False includes the napari viewer frame in the screenshot,
            By default, True.

        Returns
        -------
        image : array
            Numpy array of type ubyte and shape (h, w, 4). Index [0, 0] is the
            upper-left corner of the rendered region.
        """

        img = QImg2array(self._screenshot(size, scale, flash, canvas_only))
        if path is not None:
            imsave(path, img)
        return img

    def clipboard(self, flash=True, canvas_only=False):
        """Copy screenshot of current viewer to the clipboard.

        Parameters
        ----------
        flash : bool
            Flag to indicate whether flash animation should be shown after
            the screenshot was captured.
        canvas_only : bool
            If True, screenshot shows only the image display canvas, and
            if False include the napari viewer frame in the screenshot,
            By default, True.
        """
        img = self._screenshot(flash=flash, canvas_only=canvas_only)
        QApplication.clipboard().setImage(img)

    def _teardown(self):
        """Carry out various teardown tasks such as event disconnection."""
        self._setup_existing_themes(False)
        _themes.events.added.disconnect(self._add_theme)
        _themes.events.removed.disconnect(self._remove_theme)

    def close(self):
        """Close the viewer window and cleanup sub-widgets."""
        # Someone is closing us twice? Only try to delete self._qt_window
        # if we still have one.
        if hasattr(self, '_qt_window'):
            self._teardown()
            self._qt_viewer.close()
            self._qt_window.close()
            del self._qt_window

    def _open_preferences_dialog(self) -> PreferencesDialog:
        """Edit preferences from the menubar."""
        if self._pref_dialog is None:
            win = PreferencesDialog(parent=self._qt_window)
            self._pref_dialog = win

            app_pref = get_settings().application
            if app_pref.preferences_size:
                win.resize(*app_pref.preferences_size)

            @win.resized.connect
            def _save_size(sz: QSize):
                app_pref.preferences_size = (sz.width(), sz.height())

            def _clean_pref_dialog():
                self._pref_dialog = None

            win.finished.connect(_clean_pref_dialog)
            win.show()
        else:
            self._pref_dialog.raise_()

        return self._pref_dialog

    def _screenshot_dialog(self):
        """Save screenshot of current display with viewer, default .png"""
        from napari._qt.dialogs.screenshot_dialog import ScreenshotDialog
        from napari.utils.history import get_save_history, update_save_history

        hist = get_save_history()
        dial = ScreenshotDialog(
            self.screenshot, self._qt_viewer, hist[0], hist
        )
        if dial.exec_():
            update_save_history(dial.selectedFiles()[0])


def _instantiate_dock_widget(wdg_cls, viewer: 'Viewer'):
    # if the signature is looking a for a napari viewer, pass it.
    from napari.viewer import Viewer, ViewerModel

    kwargs = {}
    try:
        sig = inspect.signature(wdg_cls.__init__)
    # Inspection can fail when adding to bundle as it thinks widget is a builtin
    except ValueError:
        pass
    else:
        for param in sig.parameters.values():
            if param.name == 'napari_viewer':
                kwargs['napari_viewer'] = PublicOnlyProxy(viewer)
                break
            if param.annotation in (
                'napari.viewer.Viewer',
                Viewer,
                'napari.viewer.ViewerModel',
                'napari.components.ViewerModel',
                'napari.components.viewer_model.ViewerModel',
                ViewerModel,
            ):
                kwargs[param.name] = PublicOnlyProxy(viewer)
                break
            # cannot look for param.kind == param.VAR_KEYWORD because
            # QWidget allows **kwargs but errs on unknown keyword arguments

    # instantiate the widget
    return wdg_cls(**kwargs)


class InnerWidgetMappingProxy(MappingProxy):
    """A proxy for the inner widget of a QDockWidget.

    This is used to allow access to the inner widget of a QDockWidget
    without exposing the QDockWidget itself.
    """

    def __getitem__(self, key, /) -> 'QWidget | Widget':
        """Get the inner widget of the QDockWidget."""
        return self._wrapped[key].inner_widget()

    def __repr__(self) -> str:
        """Return a dict-like mapping of widget names to widget class names."""
        items = (
            f'{k!r}: {v.inner_widget()!r}' for k, v in self._wrapped.items()
        )
        if items:
            indent = '\n  '
            return (
                f'<{self.__class__.__name__} {{\n  {indent.join(items)}\n}}>'
            )
        return f'<{self.__class__.__name__} {{}}>'



from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from napari.settings import get_settings
from napari.utils.translations import trans


class ConfirmCloseDialog(QDialog):
    def __init__(self, parent, close_app=False) -> None:
        super().__init__(parent)
        cancel_btn = QPushButton(trans._('Cancel'))
        close_btn = QPushButton(trans._('Close'))
        close_btn.setObjectName('warning_icon_btn')
        icon_label = QWidget()

        self.do_not_ask = QCheckBox(trans._('Do not ask in future'))

        if close_app:
            self.setWindowTitle(trans._('Close Application?'))
            text = trans._(
                "Do you want to close the application? ('{shortcut}' to confirm). This will close all Qt Windows in this process",
                shortcut=QKeySequence('Ctrl+Q').toString(
                    QKeySequence.NativeText
                ),
            )
            close_btn.setObjectName('error_icon_btn')
            close_btn.setShortcut(QKeySequence('Ctrl+Q'))
            icon_label.setObjectName('error_icon_element')
        else:
            self.setWindowTitle(trans._('Close Window?'))
            text = trans._(
                "Confirm to close window (or press '{shortcut}')",
                shortcut=QKeySequence('Ctrl+W').toString(
                    QKeySequence.NativeText
                ),
            )
            close_btn.setObjectName('warning_icon_btn')
            close_btn.setShortcut(QKeySequence('Ctrl+W'))
            icon_label.setObjectName('warning_icon_element')

        cancel_btn.clicked.connect(self.reject)
        close_btn.clicked.connect(self.accept)

        layout = QVBoxLayout()
        layout2 = QHBoxLayout()
        layout2.addWidget(icon_label)
        layout3 = QVBoxLayout()
        layout3.addWidget(QLabel(text))
        layout3.addWidget(self.do_not_ask)
        layout2.addLayout(layout3)
        layout4 = QHBoxLayout()
        layout4.addStretch(1)
        layout4.addWidget(cancel_btn)
        layout4.addWidget(close_btn)
        layout.addLayout(layout2)
        layout.addLayout(layout4)
        self.setLayout(layout)

        # for test purposes because of problem with shortcut testing:
        # https://github.com/pytest-dev/pytest-qt/issues/254
        self.close_btn = close_btn
        self.cancel_btn = cancel_btn

    def accept(self):
        if self.do_not_ask.isChecked():
            get_settings().application.confirm_close_window = False
        super().accept()



import warnings
from collections import OrderedDict, defaultdict
from collections.abc import Iterable
from functools import lru_cache
from threading import Lock
from typing import NamedTuple, Union

import numpy as np
import skimage.color as colorconv
from vispy.color import (
    Color,
    ColorArray,
    Colormap as VispyColormap,
    get_colormap,
    get_colormaps,
)
from vispy.color.colormap import LUT_len

from napari.utils.colormaps._accelerated_cmap import minimum_dtype_for_labels
from napari.utils.colormaps.bop_colors import bopd
from napari.utils.colormaps.colormap import (
    Colormap,
    ColormapInterpolationMode,
    CyclicLabelColormap,
    DirectLabelColormap,
)
from napari.utils.colormaps.inverse_colormaps import inverse_cmaps
from napari.utils.colormaps.standardize_color import transform_color
from napari.utils.colormaps.vendored.cm import cmap_d
from napari.utils.translations import trans

# All parsable input color types that a user can provide
ColorType = Union[list, tuple, np.ndarray, str, Color, ColorArray]


ValidColormapArg = Union[
    str,
    ColorType,
    VispyColormap,
    Colormap,
    tuple[str, VispyColormap],
    tuple[str, Colormap],
    dict[str, VispyColormap],
    dict[str, Colormap],
    dict,
]


matplotlib_colormaps = _MATPLOTLIB_COLORMAP_NAMES = OrderedDict(
    viridis=trans._p('colormap', 'viridis'),
    magma=trans._p('colormap', 'magma'),
    inferno=trans._p('colormap', 'inferno'),
    plasma=trans._p('colormap', 'plasma'),
    hsv=trans._p('colormap', 'hsv'),
    turbo=trans._p('colormap', 'turbo'),
    twilight=trans._p('colormap', 'twilight'),
    twilight_shifted=trans._p('colormap', 'twilight shifted'),
    gist_earth=trans._p('colormap', 'gist earth'),
    PiYG=trans._p('colormap', 'PiYG'),
)
_MATPLOTLIB_COLORMAP_NAMES_REVERSE = {
    v: k for k, v in matplotlib_colormaps.items()
}

# some colormaps use BaseColormap and custom mapping functions instead of
# standard colors/controls, so they are broken in napari
_VISPY_COLORMAPS_ORIGINAL = _VCO = {
    k: v for k, v in get_colormaps().items() if isinstance(v, VispyColormap)
}
_VISPY_COLORMAPS_TRANSLATIONS = OrderedDict(
    autumn=(trans._p('colormap', 'autumn'), _VCO['autumn']),
    blues=(trans._p('colormap', 'blues'), _VCO['blues']),
    cool=(trans._p('colormap', 'cool'), _VCO['cool']),
    greens=(trans._p('colormap', 'greens'), _VCO['greens']),
    reds=(trans._p('colormap', 'reds'), _VCO['reds']),
    spring=(trans._p('colormap', 'spring'), _VCO['spring']),
    summer=(trans._p('colormap', 'summer'), _VCO['summer']),
    light_blues=(trans._p('colormap', 'light blues'), _VCO['light_blues']),
    orange=(trans._p('colormap', 'orange'), _VCO['orange']),
    viridis=(trans._p('colormap', 'viridis'), _VCO['viridis']),
    coolwarm=(trans._p('colormap', 'coolwarm'), _VCO['coolwarm']),
    PuGr=(trans._p('colormap', 'PuGr'), _VCO['PuGr']),
    GrBu=(trans._p('colormap', 'GrBu'), _VCO['GrBu']),
    GrBu_d=(trans._p('colormap', 'GrBu_d'), _VCO['GrBu_d']),
    RdBu=(trans._p('colormap', 'RdBu'), _VCO['RdBu']),
    cubehelix=(trans._p('colormap', 'cubehelix'), _VCO['cubehelix']),
    single_hue=(trans._p('colormap', 'single hue'), _VCO['single_hue']),
    hsl=(trans._p('colormap', 'hsl'), _VCO['hsl']),
    husl=(trans._p('colormap', 'husl'), _VCO['husl']),
    diverging=(trans._p('colormap', 'diverging'), _VCO['diverging']),
    RdYeBuCy=(trans._p('colormap', 'RdYeBuCy'), _VCO['RdYeBuCy']),
)
_VISPY_COLORMAPS_TRANSLATIONS_REVERSE = {
    v[0]: k for k, v in _VISPY_COLORMAPS_TRANSLATIONS.items()
}
_PRIMARY_COLORS = OrderedDict(
    red=(trans._p('colormap', 'red'), [1.0, 0.0, 0.0]),
    green=(trans._p('colormap', 'green'), [0.0, 1.0, 0.0]),
    blue=(trans._p('colormap', 'blue'), [0.0, 0.0, 1.0]),
    cyan=(trans._p('colormap', 'cyan'), [0.0, 1.0, 1.0]),
    magenta=(trans._p('colormap', 'magenta'), [1.0, 0.0, 1.0]),
    yellow=(trans._p('colormap', 'yellow'), [1.0, 1.0, 0.0]),
    gray=(trans._p('colormap', 'gray'), [1.0, 1.0, 1.0]),
)

SIMPLE_COLORMAPS = {
    name: Colormap(
        name=name, display_name=display_name, colors=[[0.0, 0.0, 0.0], color]
    )
    for name, (display_name, color) in _PRIMARY_COLORS.items()
}

# readd fire and ice colormap for backwards compatibility (see #7858)
VISPY_OLD_COLORMAPS = {
    'fire': Colormap(
        name='fire',
        display_name='fire',
        colors=[
            [1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 0.0, 1.0],
            [1.0, 0.0, 0.0, 1.0],
        ],
    ),
    'ice': Colormap(
        name='ice',
        display_name='ice',
        colors=[[0.0, 0.0, 1.0, 1.0], [1.0, 1.0, 1.0, 1.0]],
    ),
}

# dictionary for bop colormap objects
BOP_COLORMAPS = {
    name: Colormap(value, name=name, display_name=display_name)
    for name, (display_name, value) in bopd.items()
}

INVERSE_COLORMAPS = {
    name: Colormap(value, name=name, display_name=display_name)
    for name, (display_name, value) in inverse_cmaps.items()
}

# Add the reversed grayscale colormap (white to black) to inverse colormaps
INVERSE_COLORMAPS.update(
    {
        'gray_r': Colormap(
            name='gray_r',
            display_name='gray r',
            colors=[[1.0, 1.0, 1.0], [0.0, 0.0, 0.0]],
        ),
    }
)

DISCONTINUOUS_COLORMAPS = {
    'HiLo': Colormap(
        name='HiLo',
        display_name='HiLo',
        colors=[[0.0, 0.0, 0.0, 1.0], [1.0, 1.0, 1.0, 1.0]],
        high_color=[1.0, 0.0, 0.0, 1.0],
        low_color=[0.0, 0.0, 1.0, 1.0],
    ),
    'nan': Colormap(
        name='nan',
        display_name='NaN',
        colors=[[0.0, 0.0, 0.0, 1.0], [1.0, 1.0, 1.0, 1.0]],
        bad_color=[1.0, 0.0, 0.0, 1.0],
    ),
}

_FLOAT32_MAX = float(np.finfo(np.float32).max)
_MAX_VISPY_SUPPORTED_VALUE = _FLOAT32_MAX / 8
# Using 8 as divisor comes from experiments.
# For some reason if use smaller number,
# the image is not displayed correctly.

_MINIMUM_SHADES_COUNT = 256


def _all_rgb():
    """Return all 256**3 valid rgb tuples."""
    base = np.arange(256, dtype=np.uint8)
    r, g, b = np.meshgrid(base, base, base, indexing='ij')
    return np.stack((r, g, b), axis=-1).reshape((-1, 3))


# The following values were precomputed and stored as constants
# here to avoid heavy computation when importing this module.
# The following code can be used to reproduce these values.
#
# rgb_colors = _all_rgb()
# luv_colors = colorconv.rgb2luv(rgb_colors)
# LUVMIN = np.amin(luv_colors, axis=(0,))
# LUVMAX = np.amax(luv_colors, axis=(0,))
# lab_colors = colorconv.rgb2lab(rgb_colors)
# LABMIN = np.amin(lab_colors, axis=(0,))
# LABMAX = np.amax(lab_colors, axis=(0,))

LUVMIN = np.array([0.0, -83.07790815, -134.09790293])
LUVMAX = np.array([100.0, 175.01447356, 107.39905336])
LUVRNG = LUVMAX - LUVMIN

LABMIN = np.array([0.0, -86.18302974, -107.85730021])
LABMAX = np.array([100.0, 98.23305386, 94.47812228])
LABRNG = LABMAX - LABMIN


def convert_vispy_colormap(colormap, name='vispy'):
    """Convert a vispy colormap object to a napari colormap.

    Parameters
    ----------
    colormap : vispy.color.Colormap
        Vispy colormap object that should be converted.
    name : str
        Name of colormap, optional.

    Returns
    -------
    napari.utils.Colormap
    """
    if not isinstance(colormap, VispyColormap):
        raise TypeError(
            trans._(
                'Colormap must be a vispy colormap if passed to from_vispy',
                deferred=True,
            )
        )

    if name in _VISPY_COLORMAPS_TRANSLATIONS:
        display_name, _cmap = _VISPY_COLORMAPS_TRANSLATIONS[name]
    else:
        # Unnamed colormap
        display_name = trans._(name)

    return Colormap(
        name=name,
        display_name=display_name,
        colors=colormap.colors.rgba,
        controls=colormap._controls,
        interpolation=colormap.interpolation,
        nan_color=colormap.bad_color.rgba,
        high_color=getattr(colormap.high_color, 'rgba', None),
        low_color=getattr(colormap.low_color, 'rgba', None),
    )


def _napari_cmap_to_vispy(colormap: Colormap) -> VispyColormap:
    """Convert a napari colormap to its equivalent vispy colormap."""
    cmap_args = colormap.dict()
    cmap_args.pop('name')
    cmap_args['bad_color'] = cmap_args.pop('nan_color')
    return VispyColormap(**cmap_args)


def _validate_rgb(colors, *, tolerance=0.0):
    """Return the subset of colors that is in [0, 1] for all channels.

    Parameters
    ----------
    colors : array of float, shape (N, 3)
        Input colors in RGB space.

    Returns
    -------
    filtered_colors : array of float, shape (M, 3), M <= N
        The subset of colors that are in valid RGB space.

    Other Parameters
    ----------------
    tolerance : float, optional
        Values outside of the range by less than ``tolerance`` are allowed and
        clipped to be within the range.

    Examples
    --------
    >>> colors = np.array([[  0. , 1.,  1.  ],
    ...                    [  1.1, 0., -0.03],
    ...                    [  1.2, 1.,  0.5 ]])
    >>> _validate_rgb(colors)
    array([[0., 1., 1.]])
    >>> _validate_rgb(colors, tolerance=0.15)
    array([[0., 1., 1.],
           [1., 0., 0.]])
    """
    lo = 0 - tolerance
    hi = 1 + tolerance
    valid = np.all((colors > lo) & (colors < hi), axis=1)
    filtered_colors = np.clip(colors[valid], 0, 1)
    return filtered_colors


def low_discrepancy_image(image, seed=0.5, margin=1 / 256) -> np.ndarray:
    """Generate a 1d low discrepancy sequence of coordinates.

    Parameters
    ----------
    image : array of int
        A set of labels or label image.
    seed : float
        The seed from which to start the quasirandom sequence.
        Effective range is [0,1.0), as only the decimals are used.
    margin : float
        Values too close to 0 or 1 will get mapped to the edge of the colormap,
        so we need to offset to a margin slightly inside those values. Since
        the bin size is 1/256 by default, we offset by that amount.

    Returns
    -------
    image_out : array of float
        The set of ``labels`` remapped to [0, 1] quasirandomly.

    """
    phi_mod = 0.6180339887498948482
    image_float = np.float32(image)
    image_float = seed + image_float * phi_mod
    # We now map the floats to the range [0 + margin, 1 - margin]
    image_out = margin + (1 - 2 * margin) * (
        image_float - np.floor(image_float)
    )

    # Clear zero (background) values, matching the shader behavior in _glsl_label_step
    image_out[image == 0] = 0.0
    return image_out


def color_dict_to_colormap(colors):
    """
    Generate a color map based on the given color dictionary

    Parameters
    ----------
    colors : dict of int to array of float, shape (4)
        Mapping between labels and color

    Returns
    -------
    colormap : napari.utils.Colormap
        Colormap constructed with provided control colors
    label_color_index : dict of int
        Mapping of Label to color control point within colormap
    """

    MAX_DISTINCT_COLORS = LUT_len

    control_colors = np.unique(list(colors.values()), axis=0)

    if len(control_colors) >= MAX_DISTINCT_COLORS:
        warnings.warn(
            trans._(
                'Label layers with more than {max_distinct_colors} distinct colors will not render correctly. This layer has {distinct_colors}.',
                deferred=True,
                distinct_colors=str(len(control_colors)),
                max_distinct_colors=str(MAX_DISTINCT_COLORS),
            ),
            category=UserWarning,
        )

    colormap = Colormap(
        colors=control_colors, interpolation=ColormapInterpolationMode.ZERO
    )

    control2index = {
        tuple(color): control_point
        for color, control_point in zip(
            colormap.colors, colormap.controls, strict=False
        )
    }

    control_small_delta = 0.5 / len(control_colors)
    label_color_index = {
        label: np.float32(control2index[tuple(color)] + control_small_delta)
        for label, color in colors.items()
    }

    return colormap, label_color_index


def _low_discrepancy(dim, n, seed=0.5):
    """Generate a 1d, 2d, or 3d low discrepancy sequence of coordinates.

    Parameters
    ----------
    dim : one of {1, 2, 3}
        The dimensionality of the sequence.
    n : int
        How many points to generate.
    seed : float or array of float, shape (dim,)
        The seed from which to start the quasirandom sequence.
        Effective range is [0,1.0), as only the decimals are used.

    Returns
    -------
    pts : array of float, shape (n, dim)
        The sampled points.

    References
    ----------
    ..[1]: http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/  # noqa: E501
    """
    phi1 = 1.6180339887498948482
    phi2 = 1.32471795724474602596
    phi3 = 1.22074408460575947536
    seed = np.broadcast_to(seed, (1, dim))
    phi = np.array([phi1, phi2, phi3])
    g = 1 / phi
    n = np.reshape(np.arange(n), (n, 1))
    pts = (seed + (n * g[:dim])) % 1
    return pts


def _color_random(n, *, colorspace='lab', tolerance=0.0, seed=0.5):
    """Generate n random RGB colors uniformly from LAB or LUV space.

    Parameters
    ----------
    n : int
        Number of colors to generate.
    colorspace : str, one of {'lab', 'luv', 'rgb'}
        The colorspace from which to get random colors.
    tolerance : float
        How much margin to allow for out-of-range RGB values (these are
        clipped to be in-range).
    seed : float or array of float, shape (3,)
        Value from which to start the quasirandom sequence.
        Effective range is [0,1.0), as only the decimals are used.

    Returns
    -------
    rgb : array of float, shape (n, 3)
        RGB colors chosen uniformly at random from given colorspace.
    """
    factor = 6  # about 1/5 of random LUV tuples are inside the space
    expand_factor = 2
    rgb = np.zeros((0, 3))
    while len(rgb) < n:
        random = _low_discrepancy(3, n * factor, seed=seed)
        if colorspace == 'luv':
            raw_rgb = colorconv.luv2rgb(random * LUVRNG + LUVMIN)
        elif colorspace == 'rgb':
            raw_rgb = random
        else:  # 'lab' by default
            # The values in random are in [0, 1], but since the LAB colorspace
            # is not exactly contained in the unit-box, some 3-tuples might not
            # be valid LAB color coordinates. scikit-image handles this by projecting
            # such coordinates into the colorspace, but will also warn when doing this.
            with warnings.catch_warnings():
                # skimage 0.20.0rc
                warnings.filterwarnings(
                    action='ignore',
                    message='Conversion from CIE-LAB, via XYZ to sRGB color space resulted in',
                    category=UserWarning,
                )
                # skimage <0.20
                warnings.filterwarnings(
                    action='ignore',
                    message='Color data out of range',
                    category=UserWarning,
                )
                raw_rgb = colorconv.lab2rgb(random * LABRNG + LABMIN)
        rgb = _validate_rgb(raw_rgb, tolerance=tolerance)
        factor *= expand_factor
    return rgb[:n]


def label_colormap(
    num_colors=256, seed=0.5, background_value=0
) -> CyclicLabelColormap:
    """Produce a colormap suitable for use with a given label set.

    Parameters
    ----------
    num_colors : int, optional
        Number of unique colors to use. Default used if not given.
        Colors are in addition to a transparent color 0.
    seed : float, optional
        The seed for the random color generator.
        Effective range is [0,1.0), as only the decimals are used.

    Returns
    -------
    colormap : napari.utils.CyclicLabelColormap
        A colormap for use with labels remapped to [0, 1].

    Notes
    -----
    0 always maps to fully transparent.
    """
    if num_colors < 1:
        raise ValueError('num_colors must be >= 1')

    # Starting the control points slightly above 0 and below 1 is necessary
    # to ensure that the background pixel 0 is transparent
    midpoints = np.linspace(0.00001, 1 - 0.00001, num_colors + 1)
    control_points = np.concatenate(
        (np.array([0]), midpoints, np.array([1.0]))
    )
    # make sure to add an alpha channel to the colors

    colors = np.concatenate(
        (
            _color_random(num_colors + 2, seed=seed),
            np.full((num_colors + 2, 1), 1),
        ),
        axis=1,
    )

    # from here
    values_ = np.arange(num_colors + 2)
    randomized_values = low_discrepancy_image(values_, seed=seed)

    indices = np.clip(
        np.searchsorted(control_points, randomized_values, side='right') - 1,
        0,
        len(control_points) - 1,
    )

    # here is an ugly hack to restore classical napari color order.
    colors = colors[indices][:-1]

    # ensure that we not need to deal with differences in float rounding for
    # CPU and GPU.
    uint8_max = np.iinfo(np.uint8).max
    rgb8_colors = (colors * uint8_max).astype(np.uint8)
    colors = rgb8_colors.astype(np.float32) / uint8_max

    return CyclicLabelColormap(
        name='label_colormap',
        display_name=trans._p('colormap', 'low discrepancy colors'),
        colors=colors,
        controls=np.linspace(0, 1, len(colors) + 1),
        interpolation='zero',
        background_value=background_value,
        seed=seed,
    )


@lru_cache
def _primes(upto=2**16):
    """Generate primes up to a given number.

    Parameters
    ----------
    upto : int
        The upper limit of the primes to generate.

    Returns
    -------
    primes : np.ndarray
        The primes up to the upper limit.
    """
    primes = np.arange(3, upto + 1, 2)
    isprime = np.ones((upto - 1) // 2, dtype=bool)
    max_factor = int(np.sqrt(upto))
    for factor in primes[: max_factor // 2]:
        if isprime[(factor - 2) // 2]:
            isprime[(factor * 3 - 2) // 2 : None : factor] = 0
    return np.concatenate(([2], primes[isprime]))


def shuffle_and_extend_colormap(
    colormap: CyclicLabelColormap, seed: int, min_random_choices: int = 5
) -> CyclicLabelColormap:
    """Shuffle the colormap colors and extend it to more colors.

    The new number of colors will be a prime number that fits into the same
    dtype as the current number of colors in the colormap.

    Parameters
    ----------
    colormap : napari.utils.CyclicLabelColormap
        Colormap to shuffle and extend.
    seed : int
        Seed for the random number generator.
    min_random_choices : int
        Minimum number of new table sizes to choose from. When choosing table
        sizes, every choice gives a 1/size chance of two labels being mapped
        to the same color. Since we try to stay within the same dtype as the
        original colormap, if the number of original colors is close to the
        maximum value for the dtype, there will not be enough prime numbers
        up to the dtype max to ensure that two labels can always be
        distinguished after one or few shuffles. In that case, we discard
        some colors and choose the `min_random_choices` largest primes to fit
        within the dtype.

    Returns
    -------
    colormap : napari.utils.CyclicLabelColormap
        Shuffled and extended colormap.
    """
    rng = np.random.default_rng(seed)
    n_colors_prev = len(colormap.colors)
    dtype = minimum_dtype_for_labels(n_colors_prev)
    indices = np.arange(n_colors_prev)
    rng.shuffle(indices)
    shuffled_colors = colormap.colors[indices]

    primes = _primes(
        np.iinfo(dtype).max if np.issubdtype(dtype, np.integer) else 2**24
    )
    valid_primes = primes[primes > n_colors_prev]
    if len(valid_primes) < min_random_choices:
        valid_primes = primes[-min_random_choices:]
    n_colors = rng.choice(valid_primes)
    n_color_diff = n_colors - n_colors_prev

    extended_colors = np.concatenate(
        (
            shuffled_colors[: min(n_color_diff, 0) or None],
            shuffled_colors[rng.choice(indices, size=max(n_color_diff, 0))],
        ),
        axis=0,
    )

    new_colormap = CyclicLabelColormap(
        name=colormap.name,
        colors=extended_colors,
        controls=np.linspace(0, 1, len(extended_colors) + 1),
        interpolation='zero',
        background_value=colormap.background_value,
    )
    return new_colormap


def direct_colormap(color_dict=None):
    """Make a direct colormap from a dictionary mapping labels to colors.

    Parameters
    ----------
    color_dict : dict, optional
        A dictionary mapping labels to colors.

    Returns
    -------
    d : DirectLabelColormap
        A napari colormap whose map() function applies the color dictionary
        to an array.
    """
    # we don't actually use the color array, so pass dummy.
    return DirectLabelColormap(
        color_dict=color_dict or defaultdict(lambda: np.zeros(4)),
    )


def vispy_or_mpl_colormap(name) -> Colormap:
    """Try to get a colormap from vispy, or convert an mpl one to vispy format.

    Parameters
    ----------
    name : str
        The name of the colormap.

    Returns
    -------
    colormap : napari.utils.Colormap
        The found colormap.

    Raises
    ------
    KeyError
        If no colormap with that name is found within vispy or matplotlib.
    """
    if name in _VISPY_COLORMAPS_TRANSLATIONS:
        cmap = get_colormap(name)
        colormap = convert_vispy_colormap(cmap, name=name)
    else:
        try:
            mpl_cmap = cmap_d[name]
            display_name = _MATPLOTLIB_COLORMAP_NAMES.get(name, name)
        except KeyError as e:
            suggestion = _MATPLOTLIB_COLORMAP_NAMES_REVERSE.get(
                name
            ) or _VISPY_COLORMAPS_TRANSLATIONS_REVERSE.get(name)
            if suggestion:
                raise KeyError(
                    trans._(
                        'Colormap "{name}" not found in either vispy or matplotlib but you might want to use "{suggestion}".',
                        deferred=True,
                        name=name,
                        suggestion=suggestion,
                    )
                ) from e

            colormaps = set(_VISPY_COLORMAPS_ORIGINAL).union(set(cmap_d))
            raise KeyError(
                trans._(
                    'Colormap "{name}" not found in either vispy or matplotlib. Recognized colormaps are: {colormaps}',
                    deferred=True,
                    name=name,
                    colormaps=', '.join(sorted(f'"{cm}"' for cm in colormaps)),
                )
            ) from e
        mpl_colors = mpl_cmap(np.linspace(0, 1, 256))
        colormap = Colormap(
            name=name, display_name=display_name, colors=mpl_colors
        )

    return colormap


# A dictionary mapping names to VisPy colormap objects
ALL_COLORMAPS = {
    k: vispy_or_mpl_colormap(k) for k in _MATPLOTLIB_COLORMAP_NAMES
}
ALL_COLORMAPS.update(SIMPLE_COLORMAPS)
ALL_COLORMAPS.update(VISPY_OLD_COLORMAPS)
ALL_COLORMAPS.update(BOP_COLORMAPS)
ALL_COLORMAPS.update(INVERSE_COLORMAPS)
ALL_COLORMAPS.update(DISCONTINUOUS_COLORMAPS)

# ... sorted alphabetically by name
AVAILABLE_COLORMAPS = dict(
    sorted(ALL_COLORMAPS.items(), key=lambda cmap: cmap[0].lower())
)
# lock to allow update of AVAILABLE_COLORMAPS in threads
AVAILABLE_COLORMAPS_LOCK = Lock()

# curated colormap sets
# these are selected to look good or at least reasonable when using additive
# blending of multiple channels.
MAGENTA_GREEN = ['magenta', 'green']
RGB = ['red', 'green', 'blue']
CYMRGB = ['cyan', 'yellow', 'magenta', 'red', 'green', 'blue']


AVAILABLE_LABELS_COLORMAPS = {
    'lodisc-50': label_colormap(50),
}


def _increment_unnamed_colormap(
    existing: Iterable[str], name: str = '[unnamed colormap]'
) -> tuple[str, str]:
    """Increment name for unnamed colormap.

    NOTE: this assumes colormaps are *never* deleted, and does not check
          for name collision. If colormaps can ever be removed, please update.

    Parameters
    ----------
    existing : list of str
        Names of existing colormaps.
    name : str, optional
        Name of colormap to be incremented. by default '[unnamed colormap]'

    Returns
    -------
    name : str
        Name of colormap after incrementing.
    display_name : str
        Display name of colormap after incrementing.
    """
    display_name = trans._('[unnamed colormap]')
    if name == '[unnamed colormap]':
        past_names = [n for n in existing if n.startswith('[unnamed colormap')]
        name = f'[unnamed colormap {len(past_names)}]'
        display_name = trans._(
            '[unnamed colormap {number}]',
            number=len(past_names),
        )

    return name, display_name


def ensure_colormap(colormap: ValidColormapArg) -> Colormap:
    """Accept any valid colormap argument, and return Colormap, or raise.

    Adds any new colormaps to AVAILABLE_COLORMAPS in the process, except
    for custom unnamed colormaps created from color values.

    Parameters
    ----------
    colormap : ValidColormapArg
        See ValidColormapArg for supported input types.

    Returns
    -------
    Colormap

    Warns
    -----
    UserWarning
        If ``colormap`` is not a valid colormap argument type.

    Raises
    ------
    KeyError
        If a string is provided that is not in AVAILABLE_COLORMAPS
    TypeError
        If a tuple is provided and the first element is not a string or the
        second element is not a Colormap.
    TypeError
        If a dict is provided and any of the values are not Colormap instances
        or valid inputs to the Colormap constructor.
    """
    with AVAILABLE_COLORMAPS_LOCK:
        if isinstance(colormap, str):
            # when black given as end color, want reversed grayscale colormap
            # from white to black, named gray_r
            if colormap.startswith('#000000') or colormap.lower() == 'black':
                colormap = 'gray_r'

            # Is a colormap with this name already available?
            custom_cmap = AVAILABLE_COLORMAPS.get(colormap)
            if custom_cmap is None:
                name = (
                    colormap.lower() if colormap.startswith('#') else colormap
                )
                custom_cmap = _colormap_from_colors(colormap, name)

                if custom_cmap is None:
                    custom_cmap = vispy_or_mpl_colormap(colormap)

                for cmap_ in AVAILABLE_COLORMAPS.values():
                    if (
                        np.array_equal(cmap_.controls, custom_cmap.controls)
                        and np.array_equal(cmap_.colors, custom_cmap.colors)
                        and cmap_.interpolation == custom_cmap.interpolation
                        and np.all(cmap_.nan_color == custom_cmap.nan_color)
                        and np.all(cmap_.high_color == custom_cmap.high_color)
                        and np.all(cmap_.low_color == custom_cmap.low_color)
                    ):
                        custom_cmap = cmap_
                        break

            name = custom_cmap.name
            AVAILABLE_COLORMAPS[name] = custom_cmap
        elif isinstance(colormap, Colormap):
            AVAILABLE_COLORMAPS[colormap.name] = colormap
            name = colormap.name
        elif isinstance(colormap, VispyColormap):
            # if a vispy colormap instance is provided, make sure we don't already
            # know about it before adding a new unnamed colormap
            _name = None
            for key, val in AVAILABLE_COLORMAPS.items():
                if colormap == val:
                    _name = key
                    break

            if _name is None:
                name, _display_name = _increment_unnamed_colormap(
                    AVAILABLE_COLORMAPS
                )
            else:
                name = _name

            # Convert from vispy colormap
            cmap = convert_vispy_colormap(colormap, name=name)
            AVAILABLE_COLORMAPS[name] = cmap

        elif isinstance(colormap, tuple):
            if (
                len(colormap) == 2
                and isinstance(colormap[0], str)
                and isinstance(colormap[1], VispyColormap | Colormap)
            ):
                name = colormap[0]
                cmap = colormap[1]
                # Convert from vispy colormap
                if isinstance(cmap, VispyColormap):
                    cmap = convert_vispy_colormap(cmap, name=name)
                else:
                    cmap.name = name
                AVAILABLE_COLORMAPS[name] = cmap
            else:
                colormap = _colormap_from_colors(colormap)
                if colormap is None:
                    raise TypeError(
                        trans._(
                            'When providing a tuple as a colormap argument, either 1) the first element must be a string and the second a Colormap instance 2) or the tuple should be convertible to one or more colors',
                            deferred=True,
                        )
                    )

                name, _display_name = _increment_unnamed_colormap(
                    AVAILABLE_COLORMAPS
                )
                colormap.update({'name': name, '_display_name': _display_name})
                AVAILABLE_COLORMAPS[name] = colormap

        elif isinstance(colormap, dict):
            if 'colors' in colormap and not (
                isinstance(colormap['colors'], VispyColormap | Colormap)
            ):
                cmap = Colormap(**colormap)
                name = cmap.name
                AVAILABLE_COLORMAPS[name] = cmap
            elif not all(
                (isinstance(i, VispyColormap | Colormap))
                for i in colormap.values()
            ):
                raise TypeError(
                    trans._(
                        'When providing a dict as a colormap, all values must be Colormap instances',
                        deferred=True,
                    )
                )
            else:
                # Convert from vispy colormaps
                for key, cmap in colormap.items():
                    # Convert from vispy colormap
                    if isinstance(cmap, VispyColormap):
                        cmap = convert_vispy_colormap(cmap, name=key)
                    else:
                        cmap.name = key
                    name = key
                    colormap[name] = cmap
                AVAILABLE_COLORMAPS.update(colormap)
                if len(colormap) == 1:
                    name = next(iter(colormap))  # first key in dict
                elif len(colormap) > 1:
                    name = next(iter(colormap.keys()))

                    warnings.warn(
                        trans._(
                            'only the first item in a colormap dict is used as an argument',
                            deferred=True,
                        )
                    )
                else:
                    raise ValueError(
                        trans._(
                            'Received an empty dict as a colormap argument.',
                            deferred=True,
                        )
                    )
        else:
            colormap = _colormap_from_colors(colormap)
            if colormap is not None:
                name, _display_name = _increment_unnamed_colormap(
                    AVAILABLE_COLORMAPS
                )
                colormap.update({'name': name, '_display_name': _display_name})
                AVAILABLE_COLORMAPS[name] = colormap
            else:
                warnings.warn(
                    trans._(
                        'invalid type for colormap: {cm_type}. Must be a {{str, tuple, dict, napari.utils.Colormap, vispy.colors.Colormap}}. Reverting to default',
                        deferred=True,
                        cm_type=type(colormap),
                    )
                )
                # Use default colormap
                name = 'gray'

    return AVAILABLE_COLORMAPS[name]


def _colormap_from_colors(
    colors: ColorType,
    name: str | None = 'custom',
    display_name: str | None = None,
) -> Colormap | None:
    try:
        color_array = transform_color(colors)
    except (ValueError, AttributeError, KeyError):
        return None
    if color_array.shape[0] == 1:
        color_array = np.array([[0, 0, 0, 1], color_array[0]])
    return Colormap(color_array, name=name, display_name=display_name)


def make_default_color_array():
    return np.array([0, 0, 0, 1])


def display_name_to_name(display_name):
    display_name_map = {
        v._display_name: k for k, v in AVAILABLE_COLORMAPS.items()
    }
    return display_name_map.get(
        display_name, next(iter(AVAILABLE_COLORMAPS.keys()))
    )


class CoercedContrastLimits(NamedTuple):
    contrast_limits: tuple[float, float]
    offset: float
    scale: float

    def coerce_data(self, data: np.ndarray) -> np.ndarray:
        if self.scale <= 1:
            return data * self.scale + self.offset

        return (data + self.offset / self.scale) * self.scale


def _coerce_contrast_limits(contrast_limits: tuple[float, float]):
    """Coerce contrast limits to be in the float32 range."""
    if np.abs(contrast_limits).max() > _MAX_VISPY_SUPPORTED_VALUE:
        return scale_down(contrast_limits)

    c_min = np.float32(contrast_limits[0])
    c_max = np.float32(contrast_limits[1])
    dist = c_max - c_min
    if (
        dist < np.abs(np.spacing(c_min)) * _MINIMUM_SHADES_COUNT
        or dist < np.abs(np.spacing(c_max)) * _MINIMUM_SHADES_COUNT
    ):
        return scale_up(contrast_limits)

    return CoercedContrastLimits(contrast_limits, 0, 1)


def scale_down(contrast_limits: tuple[float, float]):
    """Scale down contrast limits to be in the float32 range."""
    scale: float = min(
        1.0,
        (_MAX_VISPY_SUPPORTED_VALUE * 2)
        / (contrast_limits[1] - contrast_limits[0]),
    )
    ctrl_lim = contrast_limits[0] * scale, contrast_limits[1] * scale
    left_shift = max(0.0, -_MAX_VISPY_SUPPORTED_VALUE - ctrl_lim[0])
    right_shift = max(0.0, ctrl_lim[1] - _MAX_VISPY_SUPPORTED_VALUE)
    offset = left_shift - right_shift
    ctrl_lim = (ctrl_lim[0] + offset, ctrl_lim[1] + offset)
    return CoercedContrastLimits(ctrl_lim, offset, scale)


def scale_up(contrast_limits: tuple[float, float]):
    """Scale up contrast limits to be in the float32 precision."""
    scale = 1000 / (contrast_limits[1] - contrast_limits[0])
    shift = -contrast_limits[0] * scale

    return CoercedContrastLimits((0, 1000), shift, scale)



from napari._vispy.overlays.axes import VispyAxesOverlay
from napari.components import ViewerModel
from napari.components.overlays import AxesOverlay


def test_init_with_2d_display_of_2_dimensions():
    axes_model = AxesOverlay()
    viewer = ViewerModel()
    viewer.dims.ndim = 2
    viewer.dims.ndisplay = 3

    axes_view = VispyAxesOverlay(viewer=viewer, overlay=axes_model)

    assert tuple(axes_view.node.text.text) == ('1', '0')


def test_init_with_2d_display_of_3_dimensions():
    axes_model = AxesOverlay()
    viewer = ViewerModel()
    viewer.dims.ndim = 3
    viewer.dims.ndisplay = 2

    axes_view = VispyAxesOverlay(viewer=viewer, overlay=axes_model)

    assert tuple(axes_view.node.text.text) == ('2', '1')


def test_init_with_3d_display_of_2_dimensions():
    axes_model = AxesOverlay()
    viewer = ViewerModel()
    viewer.dims.ndim = 2
    viewer.dims.ndisplay = 3

    axes_view = VispyAxesOverlay(viewer=viewer, overlay=axes_model)

    assert tuple(axes_view.node.text.text) == ('1', '0')


def test_init_with_3d_display_of_3_dimensions():
    axes_model = AxesOverlay()
    viewer = ViewerModel()
    viewer.dims.ndim = 3
    viewer.dims.ndisplay = 3

    axes_view = VispyAxesOverlay(viewer=viewer, overlay=axes_model)

    assert tuple(axes_view.node.text.text) == ('2', '1', '0')



import numpy as np
from vispy.visuals import VolumeVisual
from vispy.visuals.transforms.linear import STTransform

from napari._vispy.layers.image import VispyImageLayer


def vispy_image_scene_size(vispy_image: VispyImageLayer) -> np.ndarray:
    """Calculates the size of a vispy image/volume in 3D space.

    The size is the shape of the node's data multiplied by the
    node's transform scale factors.

    Returns
    -------
    np.ndarray
        The size of the node as a 3-vector of the form (x, y, z).
    """
    node = vispy_image.node
    data = node._last_data if isinstance(node, VolumeVisual) else node._data
    # Only use scale to ignore translate offset used to center top-left pixel.
    transform = STTransform(scale=np.diag(node.transform.matrix))
    # Vispy uses an xy-style ordering, whereas numpy uses a rc-style
    # ordering, so reverse the shape before applying the transform.
    size = transform.map(data.shape[::-1])
    # The last element should always be one, so ignore it.
    return size[:3]
